import pytest
from silx.gui import qt
from silx.io.dictdump import dicttoh5
import silx.io
from silx.app.view.CustomPlotSelectionWindow import CustomPlotSelectionWindow


@pytest.fixture
def silxMimeData(tmpdir):
    # Create HDF5 file with 1D dataset in tmpdir
    filename = str(tmpdir.join("test.h5"))
    dicttoh5({"x": [1, 2, 3], "y1": [1, 2], "y2": [4, 3, 2, 1], "y3": []}, filename)

    # Create mime data
    mime_data = qt.QMimeData()
    url = silx.io.url.DataUrl(file_path=filename, data_path="/x")
    mime_data.setData("application/x-silx-uri", url.path().encode())

    return mime_data


@pytest.fixture
def invalidSilxMimeData(tmpdir):
    # Create HDF5 file with 2D dataset in tmpdir
    filename = str(tmpdir.join("test.h5"))
    dicttoh5({"x": [[1, 2], [3, 4]], "y": [[1, 2], [3, 4]]}, filename)

    # Create mime data
    mime_data = qt.QMimeData()
    url = silx.io.url.DataUrl(file_path=filename, data_path="/x")
    mime_data.setData("application/x-silx-uri", url.path().encode())

    return mime_data


def testDrop(qapp, silxMimeData, qWidgetFactory):
    window = qWidgetFactory(CustomPlotSelectionWindow)

    mime_data = silxMimeData

    # Create drag enter event
    dragEnterEvent = qt.QDragEnterEvent(
        qt.QPoint(0, 0), qt.Qt.CopyAction, mime_data, qt.Qt.LeftButton, qt.Qt.NoModifier
    )
    window.getPlot1D().dragEnterEvent(dragEnterEvent)
    qapp.processEvents()

    # Create drag move event
    dragMoveEvent = qt.QDragMoveEvent(
        qt.QPoint(50, 50),
        qt.Qt.CopyAction,
        mime_data,
        qt.Qt.LeftButton,
        qt.Qt.NoModifier,
    )
    window.getPlot1D().dragMoveEvent(dragMoveEvent)
    qapp.processEvents()

    # Create drop event
    dropEvent = qt.QDropEvent(
        qt.QPointF(50, 50),
        qt.Qt.CopyAction,
        mime_data,
        qt.Qt.LeftButton,
        qt.Qt.NoModifier,
    )
    window.getPlot1D().dropEvent(dropEvent)
    qapp.processEvents()

    # Verify the drop by checking if the data is in the model
    model = window.getTreeView().model()
    assert model.rowCount() == 2

    x_item = model.getXParent()
    y_item = model.getYParent().child(0, 1)

    assert x_item is not None
    assert y_item is not None


def testRemoveDataset(silxMimeData, qapp, qWidgetFactory):
    window = qWidgetFactory(CustomPlotSelectionWindow)

    mime_data = silxMimeData

    # Create drop event for X dataset
    dragEnterEvent = qt.QDragEnterEvent(
        qt.QPoint(0, 0), qt.Qt.CopyAction, mime_data, qt.Qt.LeftButton, qt.Qt.NoModifier
    )
    window.getPlot1D().dragEnterEvent(dragEnterEvent)
    qapp.processEvents()

    # Create drop event for Y1 dataset
    dropEvent = qt.QDropEvent(
        qt.QPointF(50, 50),
        qt.Qt.CopyAction,
        mime_data,
        qt.Qt.LeftButton,
        qt.Qt.NoModifier,
    )
    window.getPlot1D().dropEvent(dropEvent)
    qapp.processEvents()

    model = window.getTreeView().model()
    assert model.getYParent().rowCount() == 2

    # Remove the dataset
    window.getTreeView()._removeFile(model.getYParent().child(0, 2), model.getYParent())
    qapp.processEvents()

    assert model.getYParent().rowCount() == 1


def testResetModel(silxMimeData, qapp, qWidgetFactory):
    window = qWidgetFactory(CustomPlotSelectionWindow)

    mime_data = silxMimeData

    # Create drop event for X dataset
    dragEnterEvent = qt.QDragEnterEvent(
        qt.QPoint(0, 0), qt.Qt.CopyAction, mime_data, qt.Qt.LeftButton, qt.Qt.NoModifier
    )
    window.getPlot1D().dragEnterEvent(dragEnterEvent)
    qapp.processEvents()

    # Create drop event for dataset
    dropEvent = qt.QDropEvent(
        qt.QPointF(50, 50),
        qt.Qt.CopyAction,
        mime_data,
        qt.Qt.LeftButton,
        qt.Qt.NoModifier,
    )
    window.getPlot1D().dropEvent(dropEvent)
    qapp.processEvents()

    model = window.getTreeView().model()
    assert model.getYParent().rowCount() == 2

    # Reset the model
    model.clearAll()
    qapp.processEvents()

    assert model.getXParent().rowCount() == 0
    assert model.getYParent().rowCount() == 1


def testMultipleDrop(qapp, silxMimeData, qWidgetFactory, qapp_utils):
    window = qWidgetFactory(CustomPlotSelectionWindow)

    mime_data = silxMimeData

    # Create drag enter event
    dragEnterEvent = qt.QDragEnterEvent(
        qt.QPoint(0, 0), qt.Qt.CopyAction, mime_data, qt.Qt.LeftButton, qt.Qt.NoModifier
    )

    # Create drag move event
    dragMoveEvent = qt.QDragMoveEvent(
        qt.QPoint(50, 50),
        qt.Qt.CopyAction,
        mime_data,
        qt.Qt.LeftButton,
        qt.Qt.NoModifier,
    )

    # Create drop event
    dropEvent = qt.QDropEvent(
        qt.QPointF(50, 50),
        qt.Qt.CopyAction,
        mime_data,
        qt.Qt.LeftButton,
        qt.Qt.NoModifier,
    )

    for _ in range(2):
        window.getPlot1D().dragEnterEvent(dragEnterEvent)
        qapp.processEvents()
        window.getPlot1D().dragMoveEvent(dragMoveEvent)
        qapp.processEvents()
        window.getPlot1D().dropEvent(dropEvent)
        qapp.processEvents()

    # Verify the drop by checking if the data is in the model
    model = window.getTreeView().model()
    assert model.rowCount() == 2
    print(model.rowCount())

    x_item = model.getXParent()
    y_item1 = model.getYParent().child(0, 1)
    y_item2 = model.getYParent().child(1, 1)

    assert x_item is not None
    assert y_item1 is not None
    assert y_item2 is not None


def testDropInvalid(qapp, invalidSilxMimeData, qWidgetFactory, qapp_utils):
    window = qWidgetFactory(CustomPlotSelectionWindow)

    mime_data = invalidSilxMimeData

    # Create drag enter event
    dragEnterEvent = qt.QDragEnterEvent(
        qt.QPoint(0, 0), qt.Qt.CopyAction, mime_data, qt.Qt.LeftButton, qt.Qt.NoModifier
    )
    window.getPlot1D().dragEnterEvent(dragEnterEvent)
    qapp.processEvents()

    # Create drag move event
    dragMoveEvent = qt.QDragMoveEvent(
        qt.QPoint(50, 50),
        qt.Qt.CopyAction,
        mime_data,
        qt.Qt.LeftButton,
        qt.Qt.NoModifier,
    )
    window.getPlot1D().dragMoveEvent(dragMoveEvent)
    qapp.processEvents()

    # Create drop event
    dropEvent = qt.QDropEvent(
        qt.QPointF(50, 50),
        qt.Qt.CopyAction,
        mime_data,
        qt.Qt.LeftButton,
        qt.Qt.NoModifier,
    )
    window.getPlot1D().dropEvent(dropEvent)
    qapp.processEvents()

    # Verify the drop by checking if the data is in the model
    model = window.getTreeView().model()
    assert model.rowCount() == 2



#!/usr/bin/env python
# /*##########################################################################
#
# Copyright (c) 2016-2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Main window used to compare images
"""

import logging
import numpy
import typing
import os.path

import silx.io
from silx.gui import icons
from silx.gui import qt
from silx.gui.plot.CompareImages import CompareImages
from silx.gui.widgets.UrlSelectionTable import UrlSelectionTable
from ..utils import parseutils
from silx.gui.plot.tools.profile.manager import ProfileManager
from silx.gui.plot.tools.compare.profile import ProfileImageDirectedLineROI

try:
    import PIL
except ImportError:
    PIL = None


_logger = logging.getLogger(__name__)


def _get_image_from_file(urlPath: str) -> typing.Optional[numpy.ndarray]:
    """Returns a dataset from an image file.

    The returned layout shape is supposed to be `rows, columns, channels (rgb[a])`.
    """
    if PIL is None:
        return None
    return numpy.asarray(PIL.Image.open(urlPath))


class CompareImagesWindow(qt.QMainWindow):
    def __init__(self, backend=None, settings=None):
        qt.QMainWindow.__init__(self, parent=None)
        self.setWindowTitle("Silx compare")

        silxIcon = icons.getQIcon("silx")
        self.setWindowIcon(silxIcon)

        self._plot = CompareImages(parent=self, backend=backend)
        self._plot.setAutoResetZoom(False)

        self.__manager = ProfileManager(self, self._plot.getPlot())
        virtualItem = self._plot._getVirtualPlotItem()
        self.__manager.setPlotItem(virtualItem)

        directedLineAction = self.__manager.createProfileAction(
            ProfileImageDirectedLineROI, self
        )

        profileToolBar = qt.QToolBar(self)
        profileToolBar.setWindowTitle("Profile")
        profileToolBar.addAction(directedLineAction)
        self.__profileToolBar = profileToolBar
        self._plot.addToolBar(profileToolBar)

        self._selectionTable = UrlSelectionTable(parent=self)
        self._selectionTable.setAcceptDrops(True)

        self.__settings = settings
        if settings:
            self.restoreSettings(settings)

        spliter = qt.QSplitter(self)
        spliter.addWidget(self._selectionTable)
        spliter.addWidget(self._plot)
        spliter.setStretchFactor(1, 1)
        spliter.setCollapsible(0, False)
        spliter.setCollapsible(1, False)
        self.__splitter = spliter

        self.setCentralWidget(spliter)

        self._selectionTable.sigImageAChanged.connect(self._updateImageA)
        self._selectionTable.sigImageBChanged.connect(self._updateImageB)

    def setUrls(self, urls):
        self.clear()
        for url in urls:
            self._selectionTable.addUrl(url)
        url1 = urls[0].path() if len(urls) >= 1 else None
        url2 = urls[1].path() if len(urls) >= 2 else None
        self._selectionTable.setUrlSelection(url_img_a=url1, url_img_b=url2)
        self._plot.resetZoom()
        self._plot.centerLines()

    def clear(self):
        self._plot.clear()
        self._selectionTable.clear()

    def _updateImageA(self, urlPath):
        try:
            data = self.readData(urlPath)
        except Exception as e:
            _logger.error("Error while loading URL %s", urlPath, exc_info=True)
            self._selectionTable.setError(urlPath, e.args[0])
            data = None
        self._plot.setImage1(data)

    def _updateImageB(self, urlPath):
        try:
            data = self.readData(urlPath)
        except Exception as e:
            _logger.error("Error while loading URL %s", urlPath, exc_info=True)
            self._selectionTable.setError(urlPath, e.args[0])
            data = None
        self._plot.setImage2(data)

    def readData(self, urlPath: str):
        """Read an URL as an image"""
        if urlPath in ("", None):
            return None

        data = None
        _, ext = os.path.splitext(urlPath)
        if ext in {".jpg", ".jpeg", ".png"}:
            try:
                data = _get_image_from_file(urlPath)
            except Exception:
                _logger.debug("Error while loading image with PIL", exc_info=True)

        if data is None:
            try:
                data = silx.io.utils.get_data(urlPath)
            except Exception:
                raise ValueError(f"Data from '{urlPath}' is not readable")

        if not isinstance(data, numpy.ndarray):
            raise ValueError(f"URL '{urlPath}' does not link to a numpy array")
        if data.dtype.kind not in set(["f", "u", "i", "b"]):
            raise ValueError(f"URL '{urlPath}' does not link to a numeric numpy array")

        if data.ndim == 2:
            return data
        if data.ndim == 3 and data.shape[2] in {3, 4}:
            return data

        raise ValueError(f"URL '{urlPath}' does not link to an numpy image")

    def closeEvent(self, event):
        settings = self.__settings
        if settings:
            self.saveSettings(self.__settings)

    def saveSettings(self, settings):
        """Save the window settings to this settings object

        :param qt.QSettings settings: Initialized settings
        """
        isFullScreen = bool(self.windowState() & qt.Qt.WindowFullScreen)
        if isFullScreen:
            # show in normal to catch the normal geometry
            self.showNormal()

        settings.beginGroup("comparewindow")
        settings.setValue("size", self.size())
        settings.setValue("pos", self.pos())
        settings.setValue("full-screen", isFullScreen)
        settings.setValue("spliter", self.__splitter.sizes())
        # NOTE: isInverted returns a numpy bool
        settings.setValue(
            "y-axis-inverted", bool(self._plot.getPlot().getYAxis().isInverted())
        )

        settings.setValue("visualization-mode", self._plot.getVisualizationMode().name)
        settings.setValue("alignment-mode", self._plot.getAlignmentMode().name)
        settings.setValue("display-keypoints", self._plot.getKeypointsVisible())

        displayKeypoints = settings.value("display-keypoints", False)
        displayKeypoints = parseutils.to_bool(displayKeypoints, False)

        # self._plot.getAlignmentMode()
        # self._plot.getVisualizationMode()
        # self._plot.getKeypointsVisible()
        settings.endGroup()

        if isFullScreen:
            self.showFullScreen()

    def restoreSettings(self, settings):
        """Restore the window settings using this settings object

        :param qt.QSettings settings: Initialized settings
        """
        settings.beginGroup("comparewindow")
        size = settings.value("size", qt.QSize(640, 480))
        pos = settings.value("pos", qt.QPoint())
        isFullScreen = settings.value("full-screen", False)
        isFullScreen = parseutils.to_bool(isFullScreen, False)
        yAxisInverted = settings.value("y-axis-inverted", False)
        yAxisInverted = parseutils.to_bool(yAxisInverted, False)

        visualizationMode = settings.value("visualization-mode", "")
        visualizationMode = parseutils.to_enum(
            visualizationMode,
            CompareImages.VisualizationMode,
            CompareImages.VisualizationMode.VERTICAL_LINE,
        )
        alignmentMode = settings.value("alignment-mode", "")
        alignmentMode = parseutils.to_enum(
            alignmentMode,
            CompareImages.AlignmentMode,
            CompareImages.AlignmentMode.ORIGIN,
        )
        displayKeypoints = settings.value("display-keypoints", False)
        displayKeypoints = parseutils.to_bool(displayKeypoints, False)

        try:
            data = settings.value("spliter")
            data = [int(d) for d in data]
            self.__splitter.setSizes(data)
        except Exception:
            _logger.debug("Backtrace", exc_info=True)
        settings.endGroup()

        if not pos.isNull():
            self.move(pos)
        if not size.isNull():
            self.resize(size)
        if isFullScreen:
            self.showFullScreen()
        self._plot.setVisualizationMode(visualizationMode)
        self._plot.setAlignmentMode(alignmentMode)
        self._plot.setKeypointsVisible(displayKeypoints)
        self._plot.getPlot().getYAxis().setInverted(yAxisInverted)



# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""
Tests for the octaveh5 module
"""

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "12/07/2016"

import unittest
import os
import tempfile

try:
    from ..octaveh5 import Octaveh5
except ImportError:
    Octaveh5 = None


@unittest.skipIf(Octaveh5 is None, "Could not import h5py")
class TestOctaveH5(unittest.TestCase):
    @staticmethod
    def _get_struct_FT():
        return {
            "NO_CHECK": 0.0,
            "SHOWSLICE": 1.0,
            "DOTOMO": 1.0,
            "DATABASE": 0.0,
            "ANGLE_OFFSET": 0.0,
            "VOLSELECTION_REMEMBER": 0.0,
            "NUM_PART": 4.0,
            "VOLOUTFILE": 0.0,
            "RINGSCORRECTION": 0.0,
            "DO_TEST_SLICE": 1.0,
            "ZEROOFFMASK": 1.0,
            "VERSION": "fastomo3 version 2.0",
            "CORRECT_SPIKES_THRESHOLD": 0.040000000000000001,
            "SHOWPROJ": 0.0,
            "HALF_ACQ": 0.0,
            "ANGLE_OFFSET_VALUE": 0.0,
            "FIXEDSLICE": "middle",
            "VOLSELECT": "total",
        }

    @staticmethod
    def _get_struct_PYHSTEXE():
        return {
            "EXE": "PyHST2_2015d",
            "VERBOSE": 0.0,
            "OFFV": "PyHST2_2015d",
            "TOMO": 0.0,
            "VERBOSE_FILE": "pyhst_out.txt",
            "DIR": "/usr/bin/",
            "OFFN": "pyhst2",
        }

    @staticmethod
    def _get_struct_FTAXIS():
        return {
            "POSITION_VALUE": 12345.0,
            "COR_ERROR": 0.0,
            "FILESDURINGSCAN": 0.0,
            "PLOTFIGURE": 1.0,
            "DIM1": 0.0,
            "OVERSAMPLING": 5.0,
            "TO_THE_CENTER": 1.0,
            "POSITION": "fixed",
            "COR_POSITION": 0.0,
            "HA": 0.0,
        }

    @staticmethod
    def _get_struct_PAGANIN():
        return {
            "MKEEP_MASK": 0.0,
            "UNSHARP_SIGMA": 0.80000000000000004,
            "DILATE": 2.0,
            "UNSHARP_COEFF": 3.0,
            "MEDIANR": 4.0,
            "DB": 500.0,
            "MKEEP_ABS": 0.0,
            "MODE": 0.0,
            "THRESHOLD": 0.5,
            "MKEEP_BONE": 0.0,
            "DB2": 100.0,
            "MKEEP_CORR": 0.0,
            "MKEEP_SOFT": 0.0,
        }

    @staticmethod
    def _get_struct_BEAMGEO():
        return {"DIST": 55.0, "SY": 0.0, "SX": 0.0, "TYPE": "p"}

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.test_3_6_fname = os.path.join(
            self.tempdir, "silx_tmp_t00_octaveTest_3_6.h5"
        )
        self.test_3_8_fname = os.path.join(
            self.tempdir, "silx_tmp_t00_octaveTest_3_8.h5"
        )

    def tearDown(self):
        if os.path.isfile(self.test_3_6_fname):
            os.unlink(self.test_3_6_fname)
        if os.path.isfile(self.test_3_8_fname):
            os.unlink(self.test_3_8_fname)

    def testWritedIsReaded(self):
        """
        Simple test to write and reaf the structure compatible with the octave h5 using structure.
        This test is for # test for octave version > 3.8
        """
        writer = Octaveh5()

        writer.open(self.test_3_8_fname, "a")
        # step 1 writing the file
        writer.write("FT", self._get_struct_FT())
        writer.write("PYHSTEXE", self._get_struct_PYHSTEXE())
        writer.write("FTAXIS", self._get_struct_FTAXIS())
        writer.write("PAGANIN", self._get_struct_PAGANIN())
        writer.write("BEAMGEO", self._get_struct_BEAMGEO())
        writer.close()

        # step 2 reading the file
        reader = Octaveh5().open(self.test_3_8_fname)
        #  2.1 check FT
        data_readed = reader.get("FT")
        self.assertEqual(data_readed, self._get_struct_FT())
        #  2.2 check PYHSTEXE
        data_readed = reader.get("PYHSTEXE")
        self.assertEqual(data_readed, self._get_struct_PYHSTEXE())
        #  2.3 check FTAXIS
        data_readed = reader.get("FTAXIS")
        self.assertEqual(data_readed, self._get_struct_FTAXIS())
        #  2.4 check PAGANIN
        data_readed = reader.get("PAGANIN")
        self.assertEqual(data_readed, self._get_struct_PAGANIN())
        #  2.5 check BEAMGEO
        data_readed = reader.get("BEAMGEO")
        self.assertEqual(data_readed, self._get_struct_BEAMGEO())
        reader.close()

    def testWritedIsReadedOldOctaveVersion(self):
        """The same test as testWritedIsReaded but for octave version < 3.8"""
        # test for octave version < 3.8
        writer = Octaveh5(3.6)

        writer.open(self.test_3_6_fname, "a")

        # step 1 writing the file
        writer.write("FT", self._get_struct_FT())
        writer.write("PYHSTEXE", self._get_struct_PYHSTEXE())
        writer.write("FTAXIS", self._get_struct_FTAXIS())
        writer.write("PAGANIN", self._get_struct_PAGANIN())
        writer.write("BEAMGEO", self._get_struct_BEAMGEO())
        writer.close()

        # step 2 reading the file
        reader = Octaveh5(3.6).open(self.test_3_6_fname)
        #  2.1 check FT
        data_readed = reader.get("FT")
        self.assertEqual(data_readed, self._get_struct_FT())
        #  2.2 check PYHSTEXE
        data_readed = reader.get("PYHSTEXE")
        self.assertEqual(data_readed, self._get_struct_PYHSTEXE())
        #  2.3 check FTAXIS
        data_readed = reader.get("FTAXIS")
        self.assertEqual(data_readed, self._get_struct_FTAXIS())
        #  2.4 check PAGANIN
        data_readed = reader.get("PAGANIN")
        self.assertEqual(data_readed, self._get_struct_PAGANIN())
        #  2.5 check BEAMGEO
        data_readed = reader.get("BEAMGEO")
        self.assertEqual(data_readed, self._get_struct_BEAMGEO())
        reader.close()



# /*##########################################################################
# Copyright (C) 2018-2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""Utils related to parsing"""

from __future__ import annotations

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "28/05/2018"

from collections.abc import Sequence
import glob
import logging
from typing import Generator, Iterable, Union, Any, Optional
from pathlib import Path


_logger = logging.getLogger(__name__)
"""Module logger"""


_trueStrings = {"yes", "true", "1"}
_falseStrings = {"no", "false", "0"}


def _string_to_bool(string: str) -> bool:
    """Returns a boolean from a string.

    :raise ValueError: If the string do not contains a boolean information.
    """
    lower = string.lower()
    if lower in _trueStrings:
        return True
    if lower in _falseStrings:
        return False
    raise ValueError("'%s' is not a valid boolean" % string)


def to_bool(thing: Any, default: Optional[bool] = None) -> bool:
    """Returns a boolean from an object.

    :raise ValueError: If the thing can't be interpreted as a boolean and
                       no default is set
    """
    if isinstance(thing, bool):
        return thing
    try:
        return _string_to_bool(thing)
    except ValueError:
        if default is not None:
            return default
        raise


def filenames_to_dataurls(
    filenames: Iterable[Union[str, Path]],
    slices: Sequence[int] = tuple(),
) -> Generator[object, None, None]:
    """Expand filenames and HDF5 data path in files input argument"""
    # Imports here so they are performed after setting HDF5_USE_FILE_LOCKING and logging level
    import silx.io
    from silx.io.utils import match
    from silx.io.url import DataUrl
    import silx.utils.files

    extra_slices = tuple(slices)

    for filename in filenames:
        url = DataUrl(filename)

        for file_path in sorted(silx.utils.files.expand_filenames([url.file_path()])):
            if url.data_path() is not None and glob.has_magic(url.data_path()):
                try:
                    with silx.io.open(file_path) as f:
                        data_paths = list(match(f, url.data_path()))
                except BaseException as e:
                    _logger.error(
                        f"Error searching HDF5 path pattern '{url.data_path()}' in file '{file_path}': Ignored"
                    )
                    _logger.error(e.args[0])
                    _logger.debug("Backtrace", exc_info=True)
                    continue
            else:
                data_paths = [url.data_path()]

            if not extra_slices:
                data_slices = (url.data_slice(),)
            elif not url.data_slice():
                data_slices = extra_slices
            else:
                data_slices = [tuple(url.data_slice()) + (s,) for s in extra_slices]

            for data_path in data_paths:
                for data_slice in data_slices:
                    yield DataUrl(
                        file_path=file_path,
                        data_path=data_path,
                        data_slice=data_slice,
                        scheme=url.scheme(),
                    )


def to_enum(thing: Any, enum_type, default: Optional[object] = None):
    """Parse this string as this enum_type."""
    try:
        v = getattr(enum_type, str(thing))
        if isinstance(v, enum_type):
            return v
        raise ValueError(f"{thing} is not a {enum_type.__name__}")
    except (AttributeError, ValueError) as e:
        if default is not None:
            return default
        raise



# /*##########################################################################
#
# Copyright (c) 2018-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module provides tool bar helper.
"""

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "09/06/2023"


import logging
import weakref
import numpy

from silx.gui import qt


_logger = logging.getLogger(__name__)


class CompareImagesStatusBar(qt.QStatusBar):
    """StatusBar containing specific information contained in a
    :class:`CompareImages` widget

    Use :meth:`setCompareWidget` to connect this toolbar to a specific
    :class:`CompareImages` widget.

    :param Union[qt.QWidget,None] parent: Parent of this widget.
    """

    def __init__(self, parent=None):
        qt.QStatusBar.__init__(self, parent)
        self.setSizeGripEnabled(False)
        self.layout().setSpacing(0)
        self.__compareWidget = None
        self._label1 = qt.QLabel(self)
        self._label1.setFrameShape(qt.QFrame.WinPanel)
        self._label1.setFrameShadow(qt.QFrame.Sunken)
        self._label2 = qt.QLabel(self)
        self._label2.setFrameShape(qt.QFrame.WinPanel)
        self._label2.setFrameShadow(qt.QFrame.Sunken)
        self._transform = qt.QLabel(self)
        self._transform.setFrameShape(qt.QFrame.WinPanel)
        self._transform.setFrameShadow(qt.QFrame.Sunken)
        self.addWidget(self._label1)
        self.addWidget(self._label2)
        self.addWidget(self._transform)
        self._pos = None
        self._updateStatusBar()

    def setCompareWidget(self, widget):
        """
        Connect this tool bar to a specific :class:`CompareImages` widget.

        :param Union[None,CompareImages] widget: The widget to connect with.
        """
        compareWidget = self.getCompareWidget()
        if compareWidget is not None:
            compareWidget.getPlot().sigPlotSignal.disconnect(self.__plotSignalReceived)
            compareWidget.sigConfigurationChanged.disconnect(self.__dataChanged)
        compareWidget = widget
        if compareWidget is None:
            self.__compareWidget = None
        else:
            self.__compareWidget = weakref.ref(compareWidget)
        if compareWidget is not None:
            compareWidget.getPlot().sigPlotSignal.connect(self.__plotSignalReceived)
            compareWidget.sigConfigurationChanged.connect(self.__dataChanged)

    def getCompareWidget(self):
        """Returns the connected widget.

        :rtype: CompareImages
        """
        if self.__compareWidget is None:
            return None
        else:
            return self.__compareWidget()

    def __plotSignalReceived(self, event):
        """Called when old style signals at emmited from the plot."""
        if event["event"] == "mouseMoved":
            x, y = event["x"], event["y"]
            self.__mouseMoved(x, y)

    def __mouseMoved(self, x, y):
        """Called when mouse move over the plot."""
        self._pos = x, y
        self._updateStatusBar()

    def __dataChanged(self):
        """Called when internal data from the connected widget changes."""
        self._updateStatusBar()

    def _formatData(self, data):
        """Format pixel of an image.

        It supports intensity, RGB, and RGBA.

        :param Union[int,float,numpy.ndarray,str]: Value of a pixel
        :rtype: str
        """
        if data is None:
            return "No data"
        if isinstance(data, (int, numpy.integer)):
            return "%d" % data
        if isinstance(data, (float, numpy.floating)):
            return "%f" % data
        if isinstance(data, numpy.ndarray):
            # RGBA value
            if data.shape == (3,):
                return "R:%d G:%d B:%d" % (data[0], data[1], data[2])
            elif data.shape == (4,):
                return "R:%d G:%d B:%d A:%d" % (data[0], data[1], data[2], data[3])
        _logger.debug("Unsupported data format %s. Cast it to string.", type(data))
        return str(data)

    def _updateStatusBar(self):
        """Update the content of the status bar"""
        widget = self.getCompareWidget()
        if widget is None:
            self._label1.setText("ImageA: NA")
            self._label2.setText("ImageB: NA")
            self._transform.setVisible(False)
        else:
            transform = widget.getTransformation()
            self._transform.setVisible(transform is not None)
            if transform is not None:
                has_notable_translation = not numpy.isclose(
                    transform.tx, 0.0, atol=0.01
                ) or not numpy.isclose(transform.ty, 0.0, atol=0.01)
                has_notable_scale = not numpy.isclose(
                    transform.sx, 1.0, atol=0.01
                ) or not numpy.isclose(transform.sy, 1.0, atol=0.01)
                has_notable_rotation = not numpy.isclose(transform.rot, 0.0, atol=0.01)

                strings = []
                if has_notable_translation:
                    strings.append("Translation")
                if has_notable_scale:
                    strings.append("Scale")
                if has_notable_rotation:
                    strings.append("Rotation")
                if strings == []:
                    has_translation = not numpy.isclose(
                        transform.tx, 0.0
                    ) or not numpy.isclose(transform.ty, 0.0)
                    has_scale = not numpy.isclose(
                        transform.sx, 1.0
                    ) or not numpy.isclose(transform.sy, 1.0)
                    has_rotation = not numpy.isclose(transform.rot, 0.0)
                    if has_translation or has_scale or has_rotation:
                        text = "No big changes"
                    else:
                        text = "No changes"
                else:
                    text = "+".join(strings)
                self._transform.setText("Align: " + text)

                strings = []
                if not numpy.isclose(transform.ty, 0.0):
                    strings.append("Translation x: %0.3fpx" % transform.tx)
                if not numpy.isclose(transform.ty, 0.0):
                    strings.append("Translation y: %0.3fpx" % transform.ty)
                if not numpy.isclose(transform.sx, 1.0):
                    strings.append("Scale x: %0.3f" % transform.sx)
                if not numpy.isclose(transform.sy, 1.0):
                    strings.append("Scale y: %0.3f" % transform.sy)
                if not numpy.isclose(transform.rot, 0.0):
                    strings.append(
                        "Rotation: %0.3fdeg" % (transform.rot * 180 / numpy.pi)
                    )
                if strings == []:
                    text = "No transformation"
                else:
                    text = "\n".join(strings)
                self._transform.setToolTip(text)

            if self._pos is None:
                self._label1.setText("ImageA: NA")
                self._label2.setText("ImageB: NA")
            else:
                data1, data2 = widget.getRawPixelData(self._pos[0], self._pos[1])
                if isinstance(data1, str):
                    self._label1.setToolTip(data1)
                    text1 = "NA"
                else:
                    self._label1.setToolTip("")
                    text1 = self._formatData(data1)
                if isinstance(data2, str):
                    self._label2.setToolTip(data2)
                    text2 = "NA"
                else:
                    self._label2.setToolTip("")
                    text2 = self._formatData(data2)
                self._label1.setText("ImageA: %s" % text1)
                self._label2.setText("ImageB: %s" % text2)



# /*##########################################################################
#
# Copyright (c) 2018-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""A widget dedicated to compare 2 images.
"""

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "23/07/2018"


import logging
import numpy
import math

import silx.image.bilinear
from silx.gui import qt
from silx.gui import plot
from silx.gui.colors import Colormap
from silx.gui.plot import tools
from silx.utils.deprecation import deprecated_warning
from silx.utils.weakref import WeakMethodProxy
from silx.gui.plot.items import Scatter
from silx.math.colormap import normalize

from .tools.compare.core import sift
from .tools.compare.core import VisualizationMode
from .tools.compare.core import AlignmentMode
from .tools.compare.core import AffineTransformation
from .tools.compare.toolbar import CompareImagesToolBar
from .tools.compare.statusbar import CompareImagesStatusBar
from .tools.compare.core import _CompareImageItem


_logger = logging.getLogger(__name__)


class CompareImages(qt.QMainWindow):
    """Widget providing tools to compare 2 images.

    .. image:: img/CompareImages.png

    :param Union[qt.QWidget,None] parent: Parent of this widget.
    :param backend: The backend to use, in:
                    'matplotlib' (default), 'mpl', 'opengl', 'gl', 'none'
                    or a :class:`BackendBase.BackendBase` class
    :type backend: str or :class:`BackendBase.BackendBase`
    """

    VisualizationMode = VisualizationMode
    """Available visualization modes"""

    AlignmentMode = AlignmentMode
    """Available alignment modes"""

    sigConfigurationChanged = qt.Signal()
    """Emitted when the configuration of the widget (visualization mode,
    alignment mode...) have changed."""

    def __init__(self, parent=None, backend=None):
        qt.QMainWindow.__init__(self, parent)
        self._resetZoomActive = True
        self._colormap = Colormap()
        """Colormap shared by all modes, except the compose images (rgb image)"""
        self._colormapKeyPoints = Colormap("spring")
        """Colormap used for sift keypoints"""

        self._colormap.sigChanged.connect(self.__colormapChanged)

        if parent is None:
            self.setWindowTitle("Compare images")
        else:
            self.setWindowFlags(qt.Qt.Widget)

        self.__transformation = None
        self.__item = _CompareImageItem()
        self.__item.setName("_virtual")
        self.__item.setColormap(self._colormap)

        self.__raw1 = None
        self.__raw2 = None
        self.__data1 = None
        self.__data2 = None
        self.__previousSeparatorPosition = None

        self.__plot = plot.PlotWidget(parent=self, backend=backend)
        self.__plot.setDefaultColormap(self._colormap)
        self.__plot.getXAxis().setLabel("Columns")
        self.__plot.getYAxis().setLabel("Rows")
        if silx.config.DEFAULT_PLOT_IMAGE_Y_AXIS_ORIENTATION == "downward":
            self.__plot.getYAxis().setInverted(True)
        self.__plot.addItem(self.__item)
        self.__plot.setActiveImage(self.__item)

        self.__plot.setKeepDataAspectRatio(True)
        self.__plot.sigPlotSignal.connect(self.__plotSlot)
        self.__plot.setAxesDisplayed(False)

        self.__scatter = Scatter()
        self.__scatter.setZValue(1)
        self.__scatter.setColormap(self._colormapKeyPoints)
        self.__plot.addItem(self.__scatter)

        self.setCentralWidget(self.__plot)

        legend = VisualizationMode.VERTICAL_LINE.name
        self.__plot.addXMarker(
            0,
            legend=legend,
            text="",
            draggable=True,
            color="blue",
            constraint=WeakMethodProxy(self.__separatorConstraint),
        )
        self.__vline = self.__plot._getMarker(legend)

        legend = VisualizationMode.HORIZONTAL_LINE.name
        self.__plot.addYMarker(
            0,
            legend=legend,
            text="",
            draggable=True,
            color="blue",
            constraint=WeakMethodProxy(self.__separatorConstraint),
        )
        self.__hline = self.__plot._getMarker(legend)

        # default values
        self.__visualizationMode = ""
        self.__alignmentMode = ""
        self.__keypointsVisible = True

        self.setAlignmentMode(AlignmentMode.ORIGIN)
        self.setVisualizationMode(VisualizationMode.VERTICAL_LINE)
        self.setKeypointsVisible(False)

        # Toolbars

        self._createToolBars(self.__plot)
        if self._interactiveModeToolBar is not None:
            self.addToolBar(self._interactiveModeToolBar)
        if self._imageToolBar is not None:
            self.addToolBar(self._imageToolBar)
        if self._compareToolBar is not None:
            self.addToolBar(self._compareToolBar)

        # Statusbar

        self._createStatusBar(self.__plot)
        if self._statusBar is not None:
            self.setStatusBar(self._statusBar)

    def __getSealedColormap(self):
        vrange = self._colormap.getColormapRange(
            self.__item.getColormappedData(copy=False)
        )
        sealed = self._colormap.copy()
        sealed.setVRange(*vrange)
        return sealed

    def __colormapChanged(self):
        sealed = self.__getSealedColormap()
        if self.__image1 is not None:
            if self.__getImageMode(self.__image1.getData(copy=False)) == "intensity":
                self.__image1.setColormap(sealed)
        if self.__image2 is not None:
            if self.__getImageMode(self.__image2.getData(copy=False)) == "intensity":
                self.__image2.setColormap(sealed)

        if "COMPOSITE" in self.__visualizationMode.name:
            self.__updateData()

    def _createStatusBar(self, plot):
        self._statusBar = CompareImagesStatusBar(self)
        self._statusBar.setCompareWidget(self)

    def _createToolBars(self, plot):
        """Create tool bars displayed by the widget"""
        toolBar = tools.InteractiveModeToolBar(parent=self, plot=plot)
        self._interactiveModeToolBar = toolBar
        toolBar = tools.ImageToolBar(parent=self, plot=plot)
        self._imageToolBar = toolBar
        toolBar = CompareImagesToolBar(self)
        toolBar.setCompareWidget(self)
        self._compareToolBar = toolBar

    def _getVirtualPlotItem(self):
        return self.__item

    def getPlot(self):
        """Returns the plot which is used to display the images.

        :rtype: silx.gui.plot.PlotWidget
        """
        return self.__plot

    def getColormap(self):
        """

        :return: colormap used for compare image
        :rtype: silx.gui.colors.Colormap
        """
        return self._colormap

    def getRawPixelData(self, x, y):
        """Return the raw pixel of each image data from axes positions.

        If the coordinate is outside of the image it returns None element in
        the tuple.

        The pixel is reach from the raw data image without filter or
        transformation. But the coordinate x and y are in the reference of the
        current displayed mode.

        :param float x: X-coordinate of the pixel in the current displayed plot
        :param float y: Y-coordinate of the pixel in the current displayed plot
        :return: A tuple of for each images containing pixel information. It
            could be a scalar value or an array in case of RGB/RGBA informations.
            It also could be a string containing information is some cases.
        :rtype: Tuple(Union[int,float,numpy.ndarray,str],Union[int,float,numpy.ndarray,str])
        """
        alignmentMode = self.__alignmentMode
        raw1, raw2 = self.__raw1, self.__raw2

        if raw1 is None or raw2 is None:
            x1 = x
            y1 = y
            x2 = x
            y2 = y
        elif alignmentMode == AlignmentMode.ORIGIN:
            x1 = x
            y1 = y
            x2 = x
            y2 = y
        elif alignmentMode == AlignmentMode.CENTER:
            yy = max(raw1.shape[0], raw2.shape[0])
            xx = max(raw1.shape[1], raw2.shape[1])
            x1 = x - (xx - raw1.shape[1]) * 0.5
            x2 = x - (xx - raw2.shape[1]) * 0.5
            y1 = y - (yy - raw1.shape[0]) * 0.5
            y2 = y - (yy - raw2.shape[0]) * 0.5
        elif alignmentMode == AlignmentMode.STRETCH:
            x1 = x
            y1 = y
            x2 = x * raw2.shape[1] / raw1.shape[1]
            y2 = x * raw2.shape[1] / raw1.shape[1]
        elif alignmentMode == AlignmentMode.AUTO:
            x1 = x
            y1 = y
            # Not implemented
            x2 = -1
            y2 = -1
        else:
            assert False

        x1, y1 = int(x1), int(y1)
        x2, y2 = int(x2), int(y2)

        if raw1 is None:
            data1 = "No image A"
        elif y1 < 0 or y1 >= raw1.shape[0] or x1 < 0 or x1 >= raw1.shape[1]:
            data1 = ""
        else:
            data1 = raw1[y1, x1]

        if raw2 is None:
            data2 = "No image B"
        elif alignmentMode == AlignmentMode.AUTO:
            data2 = "Not implemented with sift"
        elif y2 < 0 or y2 >= raw2.shape[0] or x2 < 0 or x2 >= raw2.shape[1]:
            data2 = None
        else:
            data2 = raw2[y2, x2]

        return data1, data2

    def setVisualizationMode(self, mode):
        """Set the visualization mode.

        :param str mode: New visualization to display the image comparison
        """
        if self.__visualizationMode == mode:
            return
        self.__visualizationMode = mode
        self.__item.setVizualisationMode(mode)
        self.__vline.setVisible(mode == VisualizationMode.VERTICAL_LINE)
        self.__hline.setVisible(mode == VisualizationMode.HORIZONTAL_LINE)
        self.__updateData()
        self.sigConfigurationChanged.emit()

    def centerLines(self):
        """Center the line used to compare the 2 images."""
        if self.__image1 is None:
            return
        data_range = self.__plot.getDataRange()

        if data_range[0] is not None:
            cx = (data_range[0][0] + data_range[0][1]) * 0.5
        else:
            cx = 0
        if data_range[1] is not None:
            cy = (data_range[1][0] + data_range[1][1]) * 0.5
        else:
            cy = 0
        self.__vline.setPosition(cx, cy)
        self.__hline.setPosition(cx, cy)
        self.__updateSeparators()

    def getVisualizationMode(self):
        """Returns the current interaction mode."""
        return self.__visualizationMode

    def setAlignmentMode(self, mode):
        """Set the alignment mode.

        :param str mode: New alignement to apply to images
        """
        if self.__alignmentMode == mode:
            return
        self.__alignmentMode = mode
        self.__updateData()
        self.sigConfigurationChanged.emit()

    def getAlignmentMode(self):
        """Returns the current selected alignemnt mode."""
        return self.__alignmentMode

    def getKeypointsVisible(self):
        """Returns true if the keypoints are displayed"""
        return self.__keypointsVisible

    def setKeypointsVisible(self, isVisible):
        """Set keypoints visibility.

        :param bool isVisible: If True, keypoints are displayed (if some)
        """
        if self.__keypointsVisible == isVisible:
            return
        self.__keypointsVisible = isVisible
        self.__updateKeyPoints()
        self.sigConfigurationChanged.emit()

    def __setDefaultAlignmentMode(self):
        """Reset the alignemnt mode to the default value"""
        self.setAlignmentMode(AlignmentMode.ORIGIN)

    def __plotSlot(self, event):
        """Handle events from the plot"""
        if event["event"] in ("markerMoving", "markerMoved"):
            mode = self.getVisualizationMode()
            legend = mode.name
            if event["label"] == legend:
                if mode == VisualizationMode.VERTICAL_LINE:
                    value = int(float(str(event["xdata"])))
                elif mode == VisualizationMode.HORIZONTAL_LINE:
                    value = int(float(str(event["ydata"])))
                else:
                    assert False
                if self.__previousSeparatorPosition != value:
                    self.__separatorMoved(value)
                    self.__previousSeparatorPosition = value

    def __separatorConstraint(self, x, y):
        """Manage contains on the separators to clamp them inside the images."""
        if self.__data1 is None:
            return 0, 0
        x = int(x)
        if x < 0:
            x = 0
        elif x > self.__data1.shape[1]:
            x = self.__data1.shape[1]
        y = int(y)
        if y < 0:
            y = 0
        elif y > self.__data1.shape[0]:
            y = self.__data1.shape[0]
        return x, y

    def __updateSeparators(self):
        """Redraw images according to the current state of the separators."""
        mode = self.getVisualizationMode()
        if mode == VisualizationMode.VERTICAL_LINE:
            pos = self.__vline.getXPosition()
            self.__separatorMoved(pos)
            self.__previousSeparatorPosition = pos
        elif mode == VisualizationMode.HORIZONTAL_LINE:
            pos = self.__hline.getYPosition()
            self.__separatorMoved(pos)
            self.__previousSeparatorPosition = pos
        else:
            self.__image1.setOrigin((0, 0))
            if self.__image2 is not None:
                self.__image2.setOrigin((0, 0))

    def __separatorMoved(self, pos):
        """Called when vertical or horizontal separators have moved.

        Update the displayed images.
        """
        if self.__data1 is None:
            return

        mode = self.getVisualizationMode()
        if mode == VisualizationMode.VERTICAL_LINE:
            pos = int(pos)
            if pos <= 0:
                pos = 0
            elif pos >= self.__data1.shape[1]:
                pos = self.__data1.shape[1]
            data1 = self.__data1[:, 0:pos]
            data2 = self.__data2[:, pos:]
            self.__image1.setData(data1, copy=False)
            if self.__image2 is not None:
                self.__image2.setData(data2, copy=False)
                self.__image2.setOrigin((pos, 0))
        elif mode == VisualizationMode.HORIZONTAL_LINE:
            pos = int(pos)
            if pos <= 0:
                pos = 0
            elif pos >= self.__data1.shape[0]:
                pos = self.__data1.shape[0]
            data1 = self.__data1[0:pos, :]
            data2 = self.__data2[pos:, :]
            self.__image1.setData(data1, copy=False)
            if self.__image2 is not None:
                self.__image2.setData(data2, copy=False)
                self.__image2.setOrigin((0, pos))
        else:
            assert False

    def clear(self):
        self.setData(None, None)

    def setData(self, image1, image2, updateColormap="deprecated"):
        """Set images to compare.

        Images can contains floating-point or integer values, or RGB and RGBA
        values, but should have comparable intensities.

        RGB and RGBA images are provided as an array as `[width,height,channels]`
        of unsigned integer 8-bits or floating-points between 0.0 to 1.0.

        :param numpy.ndarray image1: The first image
        :param numpy.ndarray image2: The second image
        """
        if updateColormap != "deprecated":
            deprecated_warning(
                "Argument", "setData's updateColormap argument", since_version="2.0.0"
            )

        self.__raw1 = image1
        self.__raw2 = image2
        self.__updateData()
        if self.isAutoResetZoom():
            self.__plot.resetZoom()

    def setImage1(self, image1, updateColormap="deprecated"):
        """Set image1 to be compared.

        Images can contains floating-point or integer values, or RGB and RGBA
        values, but should have comparable intensities.

        RGB and RGBA images are provided as an array as `[width,height,channels]`
        of unsigned integer 8-bits or floating-points between 0.0 to 1.0.

        :param numpy.ndarray image1: The first image
        """
        if updateColormap != "deprecated":
            deprecated_warning(
                "Argument", "setImage1's updateColormap argument", since_version="2.0.0"
            )

        self.__raw1 = image1
        self.__updateData()
        if self.isAutoResetZoom():
            self.__plot.resetZoom()

    def setImage2(self, image2, updateColormap="deprecated"):
        """Set image2 to be compared.

        Images can contains floating-point or integer values, or RGB and RGBA
        values, but should have comparable intensities.

        RGB and RGBA images are provided as an array as `[width,height,channels]`
        of unsigned integer 8-bits or floating-points between 0.0 to 1.0.

        :param numpy.ndarray image2: The second image
        """
        if updateColormap != "deprecated":
            deprecated_warning(
                "Argument", "setImage2's updateColormap argument", since_version="2.0.0"
            )

        self.__raw2 = image2
        self.__updateData()
        if self.isAutoResetZoom():
            self.__plot.resetZoom()

    def __updateKeyPoints(self):
        """Update the displayed keypoints using cached keypoints."""
        if self.__keypointsVisible and self.__matching_keypoints:
            data = self.__matching_keypoints
        else:
            data = [], [], []
        self.__scatter.setData(x=data[0], y=data[1], value=data[2])

    def __updateData(self):
        """Compute aligned image when the alignment mode changes.

        This function cache input images which are used when
        vertical/horizontal separators moves.
        """
        raw1, raw2 = self.__raw1, self.__raw2

        alignmentMode = self.getAlignmentMode()
        self.__transformation = None

        if raw1 is None or raw2 is None:
            # No need to realign the 2 images
            # But create a dummy image when there is None for simplification
            if raw1 is None:
                data1 = numpy.empty((0, 0))
            else:
                data1 = raw1
            if raw2 is None:
                data2 = numpy.empty((0, 0))
            else:
                data2 = raw2
            self.__matching_keypoints = None
        else:
            if alignmentMode == AlignmentMode.ORIGIN:
                yy = max(raw1.shape[0], raw2.shape[0])
                xx = max(raw1.shape[1], raw2.shape[1])
                size = yy, xx
                data1 = self.__createMarginImage(raw1, size, transparent=True)
                data2 = self.__createMarginImage(raw2, size, transparent=True)
                self.__matching_keypoints = [0.0], [0.0], [1.0]
            elif alignmentMode == AlignmentMode.CENTER:
                yy = max(raw1.shape[0], raw2.shape[0])
                xx = max(raw1.shape[1], raw2.shape[1])
                size = yy, xx
                data1 = self.__createMarginImage(
                    raw1, size, transparent=True, center=True
                )
                data2 = self.__createMarginImage(
                    raw2, size, transparent=True, center=True
                )
                self.__matching_keypoints = (
                    [data1.shape[1] // 2],
                    [data1.shape[0] // 2],
                    [1.0],
                )
            elif alignmentMode == AlignmentMode.STRETCH:
                data1 = raw1
                data2 = self.__rescaleImage(raw2, data1.shape)
                self.__matching_keypoints = (
                    [0, data1.shape[1], data1.shape[1], 0],
                    [0, 0, data1.shape[0], data1.shape[0]],
                    [1.0, 1.0, 1.0, 1.0],
                )
            elif alignmentMode == AlignmentMode.AUTO:
                # TODO: sift implementation do not support RGBA images
                yy = max(raw1.shape[0], raw2.shape[0])
                xx = max(raw1.shape[1], raw2.shape[1])
                size = yy, xx
                data1 = self.__createMarginImage(raw1, size)
                data2 = self.__createMarginImage(raw2, size)
                self.__matching_keypoints = [0.0], [0.0], [1.0]
                try:
                    data1, data2 = self.__createSiftData(data1, data2)
                    if data2 is None:
                        raise ValueError("Unexpected None value")
                except Exception as e:
                    # TODO: Display it on the GUI
                    _logger.error(e)
                    self.__setDefaultAlignmentMode()
                    return
            else:
                assert False

        self.__item.setImageData1(data1)
        self.__item.setImageData2(data2)

        mode = self.getVisualizationMode()
        if mode == VisualizationMode.COMPOSITE_RED_BLUE_GRAY_NEG:
            data1 = self.__composeRgbImage(data1, data2, mode)
            data2 = None
        elif mode == VisualizationMode.COMPOSITE_RED_BLUE_GRAY:
            data1 = self.__composeRgbImage(data1, data2, mode)
            data2 = None
        elif mode == VisualizationMode.COMPOSITE_A_MINUS_B:
            data1 = self.__composeAMinusBImage(data1, data2)
            data2 = None
        elif mode == VisualizationMode.ONLY_A:
            data2 = None
        elif mode == VisualizationMode.ONLY_B:
            data1 = numpy.empty((0, 0))

        self.__data1, self.__data2 = data1, data2

        colormap = self.__getSealedColormap()
        mode1 = self.__getImageMode(self.__data1)
        if mode1 == "intensity":
            colormap1 = colormap
        else:
            colormap1 = None
        self.__plot.addImage(
            data1, z=0, legend="image1", resetzoom=False, colormap=colormap1
        )
        self.__image1 = self.__plot.getImage("image1")

        if data2 is not None:
            mode2 = self.__getImageMode(data2)
            if mode2 == "intensity":
                colormap2 = colormap
            else:
                colormap2 = None
            self.__plot.addImage(
                data2, z=0, legend="image2", resetzoom=False, colormap=colormap2
            )
            self.__image2 = self.__plot.getImage("image2")
            self.__image2.setVisible(True)
        else:
            if self.__image2 is not None:
                self.__image2.setVisible(False)
            self.__image2 = None
            self.__data2 = numpy.empty((0, 0))
        self.__updateKeyPoints()

        # Set the separator into the middle
        if self.__previousSeparatorPosition is None:
            value = self.__data1.shape[1] // 2
            self.__vline.setPosition(value, 0)
            value = self.__data1.shape[0] // 2
            self.__hline.setPosition(0, value)
        self.__updateSeparators()

    def __getImageMode(self, image):
        """Returns a value identifying the way the image is stored in the
        array.

        :param numpy.ndarray image: Image to check
        :rtype: str
        """
        if len(image.shape) == 2:
            return "intensity"
        elif len(image.shape) == 3:
            if image.shape[2] == 3:
                return "rgb"
            elif image.shape[2] == 4:
                return "rgba"
        raise TypeError("'image' argument is not an image.")

    def __rescaleImage(self, image, shape):
        """Rescale an image to the requested shape.

        :rtype: numpy.ndarray
        """
        mode = self.__getImageMode(image)
        if mode == "intensity":
            data = self.__rescaleArray(image, shape)
        elif mode == "rgb":
            data = numpy.empty((shape[0], shape[1], 3), dtype=image.dtype)
            for c in range(3):
                data[:, :, c] = self.__rescaleArray(image[:, :, c], shape)
        elif mode == "rgba":
            data = numpy.empty((shape[0], shape[1], 4), dtype=image.dtype)
            for c in range(4):
                data[:, :, c] = self.__rescaleArray(image[:, :, c], shape)
        return data

    def __composeRgbImage(self, data1, data2, mode):
        """Returns an RBG image containing composition of data1 and data2 in 2
        different channels

        A data image of a size of 0 is considered as missing. This does not
        interrupt the processing.

        :param numpy.ndarray data1: First image
        :param numpy.ndarray data1: Second image
        :param VisualizationMode mode: Composition mode.
        :rtype: numpy.ndarray
        """
        if data1.size != 0 and data2.size != 0:
            assert data1.shape[0:2] == data2.shape[0:2]

        sealed = self.__getSealedColormap()
        vmin, vmax = sealed.getVRange()

        if data1.size == 0:
            intensity1 = numpy.zeros(data2.shape[0:2])
        else:
            mode1 = self.__getImageMode(data1)
            if mode1 in ["rgb", "rgba"]:
                intensity1 = self.__luminosityImage(data1)
            else:
                intensity1 = data1

        if data2.size == 0:
            intensity2 = numpy.zeros(data1.shape[0:2])
        else:
            mode2 = self.__getImageMode(data2)
            if mode2 in ["rgb", "rgba"]:
                intensity2 = self.__luminosityImage(data2)
            else:
                intensity2 = data2

        shape = intensity1.shape
        result = numpy.empty((shape[0], shape[1], 3), dtype=numpy.uint8)
        a, _, _ = normalize(
            intensity1,
            norm=sealed.getNormalization(),
            autoscale=sealed.getAutoscaleMode(),
            vmin=sealed.getVMin(),
            vmax=sealed.getVMax(),
            gamma=sealed.getGammaNormalizationParameter(),
        )
        b, _, _ = normalize(
            intensity2,
            norm=sealed.getNormalization(),
            autoscale=sealed.getAutoscaleMode(),
            vmin=sealed.getVMin(),
            vmax=sealed.getVMax(),
            gamma=sealed.getGammaNormalizationParameter(),
        )
        if mode == VisualizationMode.COMPOSITE_RED_BLUE_GRAY:
            result[:, :, 0] = a
            result[:, :, 1] = a // 2 + b // 2
            result[:, :, 2] = b
        elif mode == VisualizationMode.COMPOSITE_RED_BLUE_GRAY_NEG:
            result[:, :, 0] = 255 - b
            result[:, :, 1] = 255 - (a // 2 + b // 2)
            result[:, :, 2] = 255 - a
        return result

    def __composeAMinusBImage(self, data1, data2):
        """Returns an intensity image containing the composition of `A-B`.

        A data image of a size of 0 is considered as missing. This does not
        interrupt the processing.

        :param numpy.ndarray data1: First image
        :param numpy.ndarray data1: Second image
        :rtype: numpy.ndarray
        """
        if data1.size != 0 and data2.size != 0:
            assert data1.shape[0:2] == data2.shape[0:2]

        data1 = self.__asIntensityImage(data1)
        data2 = self.__asIntensityImage(data2)
        if data1.size == 0:
            result = data2
        elif data2.size == 0:
            result = data1
        else:
            result = data1.astype(numpy.float32) - data2.astype(numpy.float32)
        return result

    def __asIntensityImage(self, image: numpy.ndarray):
        """Returns an intensity image.

        If the image use a single channel, it will be returned as it is.

        If the image is an RBG(A) image, the luminosity (0..1) is extracted and
        returned. The alpha channel is ignored.

        :rtype: numpy.ndarray
        """
        mode = self.__getImageMode(image)
        if mode in ["rgb", "rgba"]:
            return self.__luminosityImage(image)
        return image

    def __luminosityImage(self, image: numpy.ndarray):
        """Returns the luminosity channel from an RBG(A) image.

        The alpha channel is ignored.

        :rtype: numpy.ndarray
        """
        mode = self.__getImageMode(image)
        assert mode in ["rgb", "rgba"]
        is_uint8 = image.dtype.type == numpy.uint8
        # luminosity
        image = 0.21 * image[..., 0] + 0.72 * image[..., 1] + 0.07 * image[..., 2]
        if is_uint8:
            image = image / 255.0
        return image

    def __rescaleArray(self, image, shape):
        """Rescale a 2D array to the requested shape.

        :rtype: numpy.ndarray
        """
        y, x = numpy.ogrid[: shape[0], : shape[1]]
        y, x = y * 1.0 * (image.shape[0] - 1) / (shape[0] - 1), x * 1.0 * (
            image.shape[1] - 1
        ) / (shape[1] - 1)
        b = silx.image.bilinear.BilinearImage(image)
        # TODO: could be optimized using strides
        x2d = numpy.zeros_like(y) + x
        y2d = numpy.zeros_like(x) + y
        result = b.map_coordinates((y2d, x2d))
        return result

    def __createMarginImage(self, image, size, transparent=False, center=False):
        """Returns a new image with margin to respect the requested size.

        :rtype: numpy.ndarray
        """
        assert image.shape[0] <= size[0]
        assert image.shape[1] <= size[1]
        if image.shape == size:
            return image
        mode = self.__getImageMode(image)

        if center:
            pos0 = size[0] // 2 - image.shape[0] // 2
            pos1 = size[1] // 2 - image.shape[1] // 2
        else:
            pos0, pos1 = 0, 0

        if mode == "intensity":
            data = numpy.zeros(size, dtype=image.dtype)
            data[pos0 : pos0 + image.shape[0], pos1 : pos1 + image.shape[1]] = image
            # TODO: It is maybe possible to put NaN on the margin
        else:
            if transparent:
                data = numpy.zeros((size[0], size[1], 4), dtype=numpy.uint8)
            else:
                data = numpy.zeros((size[0], size[1], 3), dtype=numpy.uint8)
            depth = min(data.shape[2], image.shape[2])
            data[
                pos0 : pos0 + image.shape[0], pos1 : pos1 + image.shape[1], 0:depth
            ] = image[:, :, 0:depth]
            if transparent and depth == 3:
                data[
                    pos0 : pos0 + image.shape[0], pos1 : pos1 + image.shape[1], 3
                ] = 255
        return data

    def __toAffineTransformation(self, sift_result):
        """Returns an affine transformation from the sift result.

        :param dict sift_result: Result of sift when using `all_result=True`
        :rtype: AffineTransformation
        """
        offset = sift_result["offset"]
        matrix = sift_result["matrix"]

        tx = offset[0]
        ty = offset[1]
        a = matrix[0, 0]
        b = matrix[0, 1]
        c = matrix[1, 0]
        d = matrix[1, 1]
        rot = math.atan2(-b, a)
        sx = (-1.0 if a < 0 else 1.0) * math.sqrt(a**2 + b**2)
        sy = (-1.0 if d < 0 else 1.0) * math.sqrt(c**2 + d**2)
        return AffineTransformation(tx, ty, sx, sy, rot)

    def getTransformation(self):
        """Returns the affine transformation applied to the second image to align
        it to the first image.

        This result is only valid for sift alignment.

        :rtype: Union[None,AffineTransformation]
        """
        return self.__transformation

    def __createSiftData(self, image, second_image):
        """Generate key points and aligned images from 2 images.

        If no keypoints matches, unaligned data are anyway returns.

        :rtype: Tuple(numpy.ndarray,numpy.ndarray)
        """
        devicetype = "GPU"

        # Compute base image
        sift_ocl = sift.SiftPlan(template=image, devicetype=devicetype)
        keypoints = sift_ocl(image)

        # Check image compatibility
        second_keypoints = sift_ocl(second_image)
        mp = sift.MatchPlan()
        match = mp(keypoints, second_keypoints)
        _logger.info("Number of Keypoints within image 1: %i" % keypoints.size)
        _logger.info("                    within image 2: %i" % second_keypoints.size)

        self.__matching_keypoints = (
            match[:].x[:, 0],
            match[:].y[:, 0],
            match[:].scale[:, 0],
        )
        matching_keypoints = match.shape[0]
        _logger.info("Matching keypoints: %i" % matching_keypoints)
        if matching_keypoints == 0:
            return image, second_image

        # TODO: Problem here is we have to compute 2 time sift
        # The first time to extract matching keypoints, second time
        # to extract the aligned image.

        # Normalize the second image
        sa = sift.LinearAlign(image, devicetype=devicetype)
        data1 = image
        # TODO: Create a sift issue: if data1 is RGB and data2 intensity
        # it returns None, while extracting manually keypoints (above) works
        result = sa.align(second_image, return_all=True)
        data2 = result["result"]
        self.__transformation = self.__toAffineTransformation(result)
        return data1, data2

    def resetZoom(self, dataMargins=None):
        """Reset the plot limits to the bounds of the data and redraw the plot."""
        self.__plot.resetZoom(dataMargins)

    def setAutoResetZoom(self, activate=True):
        """

        :param bool activate: True if we want to activate the automatic
                              plot reset zoom when setting images.
        """
        self._resetZoomActive = activate

    def isAutoResetZoom(self):
        """

        :return: True if the automatic call to resetzoom is activated
        :rtype: bool
        """
        return self._resetZoomActive



import pytest
import logging
import os
from io import BytesIO

import h5py


logger = logging.getLogger(__name__)


def _set_qt_binding(binding):
    if binding is not None:
        binding = binding.lower()
        if binding == "pyqt5":
            logger.info("Force using PyQt5")
            import PyQt5.QtCore  # noqa
        elif binding == "pyside6":
            logger.info("Force using PySide6")
            import PySide6.QtCore  # noqa
        elif binding == "pyqt6":
            logger.info("Force using PyQt6")
            import PyQt6.QtCore  # noqa
        else:
            raise ValueError("Qt binding '%s' is unknown" % binding)


def pytest_addoption(parser):
    parser.addoption(
        "--qt-binding",
        type=str,
        default=None,
        dest="qt_binding",
        help="Force using a Qt binding: 'PyQt5', 'PySide6', 'PyQt6'",
    )
    parser.addoption(
        "--no-gui",
        dest="gui",
        default=True,
        action="store_false",
        help="Disable the test of the graphical use interface",
    )
    parser.addoption(
        "--no-opengl",
        dest="opengl",
        default=True,
        action="store_false",
        help="Disable tests using OpenGL",
    )
    parser.addoption(
        "--no-opencl",
        dest="opencl",
        default=True,
        action="store_false",
        help="Disable the test of the OpenCL part",
    )
    parser.addoption(
        "--low-mem",
        dest="low_mem",
        default=False,
        action="store_true",
        help="Disable test with large memory consumption (>100Mbyte",
    )


def pytest_configure(config):
    if not config.getoption("opencl", True):
        os.environ["SILX_OPENCL"] = "False"  # Disable OpenCL support in silx

    _set_qt_binding(config.option.qt_binding)


@pytest.fixture(scope="session")
def test_options(request):
    from .test import utils

    options = utils._TestOptions()
    options.configure(request.config.option)
    yield options


@pytest.fixture(scope="class")
def test_options_class_attr(request, test_options):
    """Provides test_options as class attribute

    Used as transition from TestCase to pytest
    """
    request.cls.test_options = test_options


@pytest.fixture(scope="session")
def use_opengl(test_options):
    """Fixture to flag test using a OpenGL.

    This can be skipped with `--no-opengl`.
    """
    if not test_options.WITH_GL_TEST:
        pytest.skip(test_options.WITH_GL_TEST_REASON, allow_module_level=True)


@pytest.fixture(scope="session")
def use_opencl(test_options):
    """Fixture to flag test using a OpenCL.

    This can be skipped with `--no-opencl`.
    """
    if not test_options.WITH_OPENCL_TEST:
        pytest.skip(test_options.WITH_OPENCL_TEST_REASON, allow_module_level=True)


@pytest.fixture(scope="session")
def use_large_memory(test_options):
    """Fixture to flag test using a large memory consumption.

    This can be skipped with `--low-mem`.
    """
    if test_options.TEST_LOW_MEM:
        pytest.skip(test_options.TEST_LOW_MEM_REASON, allow_module_level=True)


@pytest.fixture(scope="session")
def use_gui(test_options):
    """Fixture to flag test using GUI.

    This can be skipped with `--no-gui`.
    """
    if not test_options.WITH_QT_TEST:
        pytest.skip(test_options.WITH_QT_TEST_REASON, allow_module_level=True)


@pytest.fixture(scope="session")
def qapp(use_gui, xvfb, request):
    _set_qt_binding(request.config.option.qt_binding)

    from silx.gui import qt

    app = qt.QApplication.instance()
    if app is None:
        app = qt.QApplication([])
    try:
        yield app
    finally:
        if app is not None:
            app.closeAllWindows()


@pytest.fixture
def qapp_utils(qapp):
    """Helper containing method to deal with QApplication and widget"""
    from silx.gui.utils.testutils import TestCaseQt

    utils = TestCaseQt()
    utils.setUpClass()
    utils.setUp()
    yield utils
    utils.tearDown()
    utils.tearDownClass()


@pytest.fixture
def qWidgetFactory(qapp, qapp_utils):
    """QWidget factory as fixture

    This fixture provides a function taking a QWidget subclass as argument
    which returns an instance of this QWidget making sure it is shown first
    and destroyed once the test is done.
    """
    from silx.gui import qt
    from silx.gui.qt.inspect import isValid

    widgets = []

    def createWidget(cls, *args, **kwargs):
        widget = cls(*args, **kwargs)
        widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        widget.show()
        qapp_utils.qWaitForWindowExposed(widget)
        widgets.append(widget)

        return widget

    yield createWidget

    qapp.processEvents()

    for widget in widgets:
        if isValid(widget):
            widget.close()
    qapp.processEvents()

    # Wait some time for all widgets to be deleted
    for _ in range(10):
        validWidgets = [widget for widget in widgets if isValid(widget)]
        if validWidgets:
            qapp_utils.qWait(10)

    validWidgets = [widget for widget in widgets if isValid(widget)]
    assert not validWidgets, f"Some widgets were not destroyed: {validWidgets}"

    del widgets


@pytest.fixture
def tmp_h5py_file():
    with BytesIO() as buffer:
        with h5py.File(buffer, mode="w") as h5file:
            yield h5file



#!/usr/bin/env python
# /*##########################################################################
#
# Copyright (c) 2016-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
This script illustrates the use of :class:`silx.gui.widgets.FloatEdit.FloatEdit`.
"""

from silx.gui import qt
from silx.gui.widgets.FloatEdit import FloatEdit


class Dialog(qt.QMainWindow):
    def __init__(self, *args, **kwargs):
        qt.QMainWindow.__init__(self, *args, **kwargs)

        widget = qt.QWidget(self)
        layout = qt.QHBoxLayout(widget)

        f1 = FloatEdit(self)
        layout.addWidget(f1)

        f2 = FloatEdit(self)
        layout.addWidget(f2)

        f3 = FloatEdit(self)
        f3.setWidgetResizable(True)
        layout.addWidget(f3)

        b = qt.QPushButton(self)
        b.setText("f3=100")
        b.clicked.connect(lambda: f3.setValue(100))
        layout.addWidget(b)

        self.setCentralWidget(widget)


if __name__ == "__main__":
    app = qt.QApplication([])
    window = Dialog()
    window.setVisible(True)
    app.exec()



# /*##########################################################################
#
# Copyright (c) 2016-2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Set of icons for buttons.

Use :func:`getQIcon` to create Qt QIcon from the name identifying an icon.
"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "07/01/2019"


import os
import logging
import weakref
from . import qt
import silx.resources
from silx.utils import weakref as silxweakref


_logger = logging.getLogger(__name__)
"""Module logger"""


_cached_icons = None
"""Cache loaded icons in a weak structure"""


def getIconCache():
    """Get access to all cached icons

    :rtype: dict
    """
    global _cached_icons
    if _cached_icons is None:
        _cached_icons = weakref.WeakValueDictionary()
        # Clean up the cache before leaving the application
        # See https://github.com/silx-kit/silx/issues/1771
        qt.QApplication.instance().aboutToQuit.connect(cleanIconCache)
    return _cached_icons


def cleanIconCache():
    """Clean up the icon cache"""
    _logger.debug("Clean up icon cache")
    _cached_icons.clear()


_supported_formats = None
"""Order of file format extension to check"""


class AbstractAnimatedIcon(qt.QObject):
    """Store an animated icon.

    It provides an event containing the new icon everytime it is updated."""

    def __init__(self, parent=None):
        """Constructor

        :param qt.QObject parent: Parent of the QObject
        :raises: ValueError when name is not known
        """
        qt.QObject.__init__(self, parent)

        self.__targets = silxweakref.WeakList()
        self.__currentIcon = None

    iconChanged = qt.Signal(qt.QIcon)
    """Signal sent with a QIcon everytime the animation changed."""

    def register(self, obj):
        """Register an object to the AbstractAnimatedIcon.
        If no object are registred, the animation is paused.
        Object are stored in a weaked list.

        :param object obj: An object
        """
        if obj not in self.__targets:
            self.__targets.append(obj)
        self._updateState()

    def unregister(self, obj):
        """Remove the object from the registration.
        If no object are registred the animation is paused.

        :param object obj: A registered object
        """
        if obj in self.__targets:
            self.__targets.remove(obj)
        self._updateState()

    def hasRegistredObjects(self):
        """Returns true if any object is registred.

        :rtype: bool
        """
        return len(self.__targets)

    def isRegistered(self, obj):
        """Returns true if the object is registred in the AbstractAnimatedIcon.

        :param object obj: An object
        :rtype: bool
        """
        return obj in self.__targets

    def currentIcon(self):
        """Returns the icon of the current frame.

        :rtype: qt.QIcon
        """
        return self.__currentIcon

    def _updateState(self):
        """Update the object according to the connected objects."""
        pass

    def _setCurrentIcon(self, icon):
        """Store the current icon and emit a `iconChanged` event.

        :param qt.QIcon icon: The current icon
        """
        self.__currentIcon = icon
        self.iconChanged.emit(self.__currentIcon)


class MovieAnimatedIcon(AbstractAnimatedIcon):
    """Store a looping QMovie to provide icons for each frames.
    Provides an event with the new icon everytime the movie frame
    is updated."""

    def __init__(self, filename, parent=None):
        """Constructor

        :param str filename: An icon name to an animated format
        :param qt.QObject parent: Parent of the QObject
        :raises: ValueError when name is not known
        """
        AbstractAnimatedIcon.__init__(self, parent)

        qfile = getQFile(filename)
        self.__movie = qt.QMovie(qfile.fileName(), qt.QByteArray(), parent)
        self.__movie.setCacheMode(qt.QMovie.CacheAll)
        self.__movie.frameChanged.connect(self.__frameChanged)
        self.__cacheIcons = {}

        self.__movie.jumpToFrame(0)
        self.__updateIconAtFrame(0)

    def __frameChanged(self, frameId):
        """Callback everytime the QMovie frame change
        :param int frameId: Current frame id
        """
        self.__updateIconAtFrame(frameId)

    def __updateIconAtFrame(self, frameId):
        """
        Update the current stored QIcon

        :param int frameId: Current frame id
        """
        if frameId in self.__cacheIcons:
            icon = self.__cacheIcons[frameId]
        else:
            icon = qt.QIcon(self.__movie.currentPixmap())
            self.__cacheIcons[frameId] = icon
        self._setCurrentIcon(icon)

    def _updateState(self):
        """Update the movie play according to internal stat of the
        MovieAnimatedIcon."""
        self.__movie.setPaused(not self.hasRegistredObjects())


class MultiImageAnimatedIcon(AbstractAnimatedIcon):
    """Store a looping QMovie to provide icons for each frames.
    Provides an event with the new icon everytime the movie frame
    is updated."""

    def __init__(self, filename, parent=None):
        """Constructor

        :param str filename: An icon name to an animated format
        :param qt.QObject parent: Parent of the QObject
        :raises: ValueError when name is not known
        """
        AbstractAnimatedIcon.__init__(self, parent)

        self.__frames = []
        for i in range(100):
            try:
                frame_filename = os.sep.join((filename, ("%02d" % i)))
                frame_file = getQFile(frame_filename)
            except ValueError:
                break
            try:
                icon = qt.QIcon(frame_file.fileName())
            except ValueError:
                break
            self.__frames.append(icon)

        if len(self.__frames) == 0:
            raise ValueError("Animated icon '%s' do not exists" % filename)

        self.__frameId = -1
        self.__timer = qt.QTimer(self)
        self.__timer.timeout.connect(self.__increaseFrame)
        self.__updateIconAtFrame(0)

    def __increaseFrame(self):
        """Callback called every timer timeout to change the current frame of
        the animation
        """
        frameId = (self.__frameId + 1) % len(self.__frames)
        self.__updateIconAtFrame(frameId)

    def __updateIconAtFrame(self, frameId):
        """
        Update the current stored QIcon

        :param int frameId: Current frame id
        """
        self.__frameId = frameId
        icon = self.__frames[frameId]
        self._setCurrentIcon(icon)

    def _updateState(self):
        """Update the object to wake up or sleep it according to its use."""
        if self.hasRegistredObjects():
            if not self.__timer.isActive():
                self.__timer.start(100)
        else:
            if self.__timer.isActive():
                self.__timer.stop()


def getWaitIcon():
    """Returns a cached version of the waiting AbstractAnimatedIcon.

    :rtype: AbstractAnimatedIcon
    """
    return getAnimatedIcon("process-working")


def getAnimatedIcon(name):
    """Create an AbstractAnimatedIcon from a resource name.

    The resource name can be prefixed by the name of a resource directory. For
    example "silx:foo.png" identify the resource "foo.png" from the resource
    directory "silx".

    If no prefix are specified, the file with be returned from the silx
    resource directory with a specific path "gui/icons".

    See also :func:`silx.resources.register_resource_directory`.

    Try to load a mng or a gif file, then try to load a multi-image animated
    icon.

    In Qt5 mng or gif are not used, because the transparency is not very well
    managed.

    :param str name: Name of the icon, in one of the defined icons
                     in this module.
    :return: Corresponding AbstractAnimatedIcon
    :raises: ValueError when name is not known
    """
    key = name + "__anim"
    cached_icons = getIconCache()
    if key not in cached_icons:
        qtMajorVersion = int(qt.qVersion().split(".")[0])
        icon = None

        # ignore mng and gif in Qt5
        if qtMajorVersion != 5:
            try:
                icon = MovieAnimatedIcon(name)
            except ValueError:
                icon = None

        if icon is None:
            try:
                icon = MultiImageAnimatedIcon(name)
            except ValueError:
                icon = None

        if icon is None:
            raise ValueError("Not an animated icon name: %s", name)

        cached_icons[key] = icon
    else:
        icon = cached_icons[key]
    return icon


def getQIcon(name):
    """Create a QIcon from its name.

    The resource name can be prefixed by the name of a resource directory. For
    example "silx:foo.png" identify the resource "foo.png" from the resource
    directory "silx".

    If no prefix are specified, the file with be returned from the silx
    resource directory with a specific path "gui/icons".

    See also :func:`silx.resources.register_resource_directory`.

    :param str name: Name of the icon, in one of the defined icons
                     in this module.
    :return: Corresponding QIcon
    :raises: ValueError when name is not known
    """
    cached_icons = getIconCache()
    if name not in cached_icons:
        qfile = getQFile(name)
        icon = qt.QIcon(qfile.fileName())
        cached_icons[name] = icon
    else:
        icon = cached_icons[name]
    return icon


def getQPixmap(name):
    """Create a QPixmap from its name.

    The resource name can be prefixed by the name of a resource directory. For
    example "silx:foo.png" identify the resource "foo.png" from the resource
    directory "silx".

    If no prefix are specified, the file with be returned from the silx
    resource directory with a specific path "gui/icons".

    See also :func:`silx.resources.register_resource_directory`.

    :param str name: Name of the icon, in one of the defined icons
                     in this module.
    :return: Corresponding QPixmap
    :raises: ValueError when name is not known
    """
    qfile = getQFile(name)
    return qt.QPixmap(qfile.fileName())


def getQFile(name):
    """Create a QFile from an icon name. Filename is found
    according to supported Qt formats.

    The resource name can be prefixed by the name of a resource directory. For
    example "silx:foo.png" identify the resource "foo.png" from the resource
    directory "silx".

    If no prefix are specified, the file with be returned from the silx
    resource directory with a specific path "gui/icons".

    See also :func:`silx.resources.register_resource_directory`.

    :param str name: Name of the icon, in one of the defined icons
                     in this module.
    :return: Corresponding QFile
    :rtype: qt.QFile
    :raises: ValueError when name is not known
    """
    global _supported_formats
    if _supported_formats is None:
        _supported_formats = []
        supported_formats = qt.supportedImageFormats()
        order = ["mng", "gif", "svg", "png", "jpg"]
        for format_ in order:
            if format_ in supported_formats:
                _supported_formats.append(format_)
        if len(_supported_formats) == 0:
            _logger.error("No format supported for icons")
        else:
            _logger.debug("Format %s supported", ", ".join(_supported_formats))

    for format_ in _supported_formats:
        format_ = str(format_)
        filename = silx.resources._resource_filename(
            "%s.%s" % (name, format_), default_directory="gui/icons"
        )
        qfile = qt.QFile(filename)
        if qfile.exists():
            return qfile
        _logger.debug("File '%s' not found.", filename)
    raise ValueError("Not an icon name: %s" % name)



# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""
histogramnd benchmarks, vs numpy.histogramdd (bin counts and weights).
"""

import numpy as np

import time

from silx.math import histogramnd


def print_times(t0s, t1s, t2s, t3s):
    c_times = t1s - t0s
    np_times = t2s - t1s
    np_w_times = t3s - t2s

    time_txt = "min : {0: <7.3f}; max : {1: <7.3f}; avg : {2: <7.3f}"

    print("\tTimes :")
    print("\tC     : " + time_txt.format(c_times.min(), c_times.max(), c_times.mean()))
    print(
        "\tNP    : " + time_txt.format(np_times.min(), np_times.max(), np_times.mean())
    )
    print(
        "\tNP(W) : "
        + time_txt.format(np_w_times.min(), np_w_times.max(), np_w_times.mean())
    )


def commpare_results(
    txt, times, result_c, result_np, result_np_w, sample, weights, raise_ex=False
):
    if result_np:
        hits_cmp = np.array_equal(result_c[0], result_np[0])
    else:
        hits_cmp = None

    if result_np_w and result_c[1] is not None:
        weights_cmp = np.array_equal(result_c[1], result_np_w[0])
    else:
        weights_cmp = None

    if (hits_cmp is not None and not hits_cmp) or (
        weights_cmp is not None and not weights_cmp
    ):
        err_txt = (
            txt + " : results arent the same : "
            "hits : {0}, "
            "weights : {1}."
            "".format("OK" if hits_cmp else "NOK", "OK" if weights_cmp else "NOK")
        )
        print("\t" + err_txt)
        if raise_ex:
            raise ValueError(err_txt)
        return False

    result_txt = " : results OK. c : {0: <7.3f};".format(times[0])
    if result_np or result_np_w:
        result_txt += (
            " np : {0: <7.3f}; "
            "np (weights) {1: <7.3f}."
            "".format(times[1], times[2])
        )
    print("\t" + txt + result_txt)
    return True


def benchmark(
    n_loops,
    sample_shape,
    sample_rng,
    weights_rng,
    histo_range,
    n_bins,
    weight_min,
    weight_max,
    last_bin_closed,
    dtype=np.double,
    do_weights=True,
    do_numpy=True,
):
    int_min = 0
    int_max = 100000

    sample = np.random.randint(int_min, high=int_max, size=sample_shape).astype(
        np.double
    )
    sample = sample_rng[0] + (sample - int_min) * (sample_rng[1] - sample_rng[0]) / (
        int_max - int_min
    )
    sample = sample.astype(dtype)

    if do_weights:
        weights = np.random.randint(
            int_min, high=int_max, size=(ssetup.pyample_shape[0],)
        )
        weights = weights.astype(np.double)
        weights = weights_rng[0] + (weights - int_min) * (
            weights_rng[1] - weights_rng[0]
        ) / (int_max - int_min)
    else:
        weights = None

    t0s = []
    t1s = []
    t2s = []
    t3s = []

    for i in range(n_loops):
        t0s.append(time.time())
        result_c = histogramnd(
            sample,
            histo_range,
            n_bins,
            weights=weights,
            weight_min=weight_min,
            weight_max=weight_max,
            last_bin_closed=last_bin_closed,
        )
        t1s.append(time.time())
        if do_numpy:
            result_np = np.histogramdd(sample, bins=n_bins, range=histo_range)
            t2s.append(time.time())
            result_np_w = np.histogramdd(
                sample, bins=n_bins, range=histo_range, weights=weights
            )
            t3s.append(time.time())
        else:
            result_np = None
            result_np_w = None
            t2s.append(0)
            t3s.append(0)

        commpare_results(
            "Run {0}".format(i),
            [t1s[-1] - t0s[-1], t2s[-1] - t1s[-1], t3s[-1] - t2s[-1]],
            result_c,
            result_np,
            result_np_w,
            sample,
            weights,
        )

    print_times(np.array(t0s), np.array(t1s), np.array(t2s), np.array(t3s))


def run_benchmark(dtype=np.double, do_weights=True, do_numpy=True):
    n_loops = 5

    weights_rng = [0.0, 100.0]
    sample_rng = [0.0, 100.0]

    weight_min = None
    weight_max = None
    last_bin_closed = True

    # ====================================================
    # ====================================================
    # 1D
    # ====================================================
    # ====================================================

    print("==========================")
    print(" 1D [{0}]".format(dtype))
    print("==========================")
    sample_shape = (10**7,)
    histo_range = [[0.0, 100.0]]
    n_bins = 30

    benchmark(
        n_loops,
        sample_shape,
        sample_rng,
        weights_rng,
        histo_range,
        n_bins,
        weight_min,
        weight_max,
        last_bin_closed,
        dtype=dtype,
        do_weights=True,
        do_numpy=do_numpy,
    )

    # ====================================================
    # ====================================================
    # 2D
    # ====================================================
    # ====================================================

    print("==========================")
    print(" 2D [{0}]".format(dtype))
    print("==========================")
    sample_shape = (10**7, 2)
    histo_range = [[0.0, 100.0], [0.0, 100.0]]
    n_bins = 30

    benchmark(
        n_loops,
        sample_shape,
        sample_rng,
        weights_rng,
        histo_range,
        n_bins,
        weight_min,
        weight_max,
        last_bin_closed,
        dtype=dtype,
        do_weights=True,
        do_numpy=do_numpy,
    )

    # ====================================================
    # ====================================================
    # 3D
    # ====================================================
    # ====================================================

    print("==========================")
    print(" 3D [{0}]".format(dtype))
    print("==========================")
    sample_shape = (10**7, 3)
    histo_range = np.array([[0.0, 100.0], [0.0, 100.0], [0.0, 100.0]])
    n_bins = 30

    benchmark(
        n_loops,
        sample_shape,
        sample_rng,
        weights_rng,
        histo_range,
        n_bins,
        weight_min,
        weight_max,
        last_bin_closed,
        dtype=dtype,
        do_weights=True,
        do_numpy=do_numpy,
    )


if __name__ == "__main__":
    types = (
        np.double,
        np.int32,
        np.float32,
    )

    for t in types:
        run_benchmark(t, do_weights=True, do_numpy=True)



#!/usr/bin/env python
# /*##########################################################################
#
# Copyright (c) 2016-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Example demonstrating the use of the widget CompareImages
"""

import sys
import logging
import numpy
import argparse
import os

import silx.io
from silx.gui import qt
import silx.test.utils
from silx.gui.plot.CompareImages import CompareImages

_logger = logging.getLogger(__name__)


def createTestData():
    data = numpy.arange(100 * 100)
    data = (data % 100) / 5.0
    data = numpy.sin(data)
    data1 = data.copy()
    data1.shape = 100, 100
    data2 = silx.test.utils.add_gaussian_noise(data, stdev=0.1)
    data2.shape = 100, 100
    return data1, data2


def createParser():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="Set logging system in debug mode",
    )
    parser.add_argument(
        "--use-opengl-plot",
        dest="use_opengl_plot",
        action="store_true",
        default=False,
        help="Use OpenGL for plots (instead of matplotlib)",
    )
    return parser


if __name__ == "__main__":
    parser = createParser()
    options = parser.parse_args(sys.argv[1:])

    if options.debug:
        logging.root.setLevel(logging.DEBUG)

    if options.use_opengl_plot:
        backend = "gl"
    else:
        backend = "mpl"

    app = qt.QApplication([])

    data1, data2 = createTestData()
    window = CompareImages(backend=backend)
    window.setData(data1, data2)
    window.setVisible(True)
    app.exec()



#
#    Project: S I L X project
#             https://github.com/silx-kit/silx
#
#    Copyright (C) 2012-2023 European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

"""
Utilities around atomic operation in OpenCL
"""

__author__ = "Jerome Kieffer"
__license__ = "MIT"
__date__ = "14/06/2023"
__copyright__ = "2023-2023, ESRF, Grenoble"
__contact__ = "jerome.kieffer@esrf.fr"

import numpy
import pyopencl
import pyopencl.array as cla


def check_atomic32(device):
    try:
        ctx = pyopencl.Context(devices=[device])
    except:
        return False, f"Unable to create context on {device}"
    else:
        queue = pyopencl.CommandQueue(ctx)
    src = """
kernel void check_atomic32(global int* ary){
int res = atom_inc(ary);
}
"""
    try:
        prg = pyopencl.Program(ctx, src).build()
    except Exception as err:
        return False, f"{type(err)}: {err}"
    a = numpy.zeros(1, numpy.int32)
    d = cla.to_device(queue, a)
    prg.check_atomic32(queue, (1024,), (32,), d.data).wait()
    value = d.get()[0]
    return value == 1024, f"Got the proper value 1024=={value}"


def check_atomic64(device):
    try:
        ctx = pyopencl.Context(devices=[device])
    except:
        return False, f"Unable to create context on {device}"
    else:
        queue = pyopencl.CommandQueue(ctx)
    if (
        device.platform.name == "Portable Computing Language"
        and "GPU" in pyopencl.device_type.to_string(device.type).upper()
    ):
        # this configuration is known to seg-fault
        return False, "PoCL + GPU do not support atomic64"
    src = """
#pragma OPENCL EXTENSION cl_khr_fp64: enable
#pragma OPENCL EXTENSION cl_khr_int64_base_atomics : enable
kernel void check_atomic64(global long* ary){
long res = atom_inc(ary);
}
"""
    try:
        prg = pyopencl.Program(ctx, src).build()
    except Exception as err:
        return False, f"{type(err)}: {err}"
    a = numpy.zeros(1, numpy.int64)
    d = cla.to_device(queue, a)
    prg.check_atomic64(queue, (1024,), (32,), d.data).wait()
    value = d.get()[0]
    return value == 1024, f"Got the proper value 1024=={value}"



# /*##########################################################################
#
# Copyright (c) 2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Test PlotWidget active item"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "13/12/2023"


import pytest
from silx.gui.plot import PlotWidget


@pytest.fixture
def plotWidget(qWidgetFactory, request):
    try:
        backend = request.param
    except AttributeError:
        backend = "mpl"  # Backend was not defined
    if backend == "gl":
        request.getfixturevalue("use_opengl")  # Skip test if OpenGL test disabled
    yield qWidgetFactory(PlotWidget, backend=backend)



# /*##########################################################################
#
# Copyright (c) 2017-2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Test of utils module."""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "16/01/2017"

import numpy
import pytest

from silx.gui import qt
from silx.gui.utils.image import convertArrayToQImage, convertQImageToArray


@pytest.mark.parametrize(
    "format_, channels",
    [
        (qt.QImage.Format_RGB888, 3),  # Native support
        (qt.QImage.Format_ARGB32, 4),  # Native support
    ],
)
def testConvertArrayToQImage(format_, channels):
    """Test conversion of numpy array to QImage"""
    image = numpy.arange(3 * 3 * channels, dtype=numpy.uint8).reshape(3, 3, channels)
    qimage = convertArrayToQImage(image)

    assert (qimage.height(), qimage.width()) == image.shape[:2]
    assert qimage.format() == format_

    for row in range(3):
        for col in range(3):
            # Qrgb has no alpha channel, not compared
            # Qt uses x,y while array is row,col...
            assert qt.QColor(qimage.pixel(col, row)) == qt.QColor(*image[row, col, :3])


@pytest.mark.parametrize(
    "format_, channels",
    [
        (qt.QImage.Format_RGB888, 3),  # Native support
        (qt.QImage.Format_ARGB32, 4),  # Native support
        (qt.QImage.Format_RGB32, 3),  # Conversion to RGB
    ],
)
def testConvertQImageToArray(format_, channels):
    """Test conversion of QImage to numpy array"""
    color = numpy.arange(channels)  # RGB(A) values
    qimage = qt.QImage(3, 3, format_)
    qimage.fill(qt.QColor(*color))
    image = convertQImageToArray(qimage)

    assert (qimage.height(), qimage.width(), len(color)) == image.shape
    assert numpy.all(numpy.equal(image, color))


def testConvertQImageToArrayGrayscale():
    """Test conversion of grayscale QImage to numpy array"""
    qimage = qt.QImage(3, 3, qt.QImage.Format_Grayscale8)
    qimage.fill(1)
    image = convertQImageToArray(qimage)

    assert (qimage.height(), qimage.width()) == image.shape
    assert numpy.all(numpy.equal(image, 1))



# /*##########################################################################
#
# Copyright (c) 2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Tests of PlotWidget Axis items"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "15/06/2023"


from silx.gui.plot import PlotWidget


def testAxisIsVisible(qapp, qWidgetFactory):
    """Test Axis.isVisible method"""
    plotWidget = qWidgetFactory(PlotWidget)

    assert plotWidget.getXAxis().isVisible()
    assert plotWidget.getYAxis().isVisible()
    assert not plotWidget.getYAxis("right").isVisible()

    # Add curve on right axis
    plotWidget.addCurve((0, 1, 2), (1, 2, 3), yaxis="right")
    qapp.processEvents()

    assert plotWidget.getYAxis("right").isVisible()

    # hide curve on right axis
    curve = plotWidget.getItems()[0]
    curve.setVisible(False)
    qapp.processEvents()

    assert not plotWidget.getYAxis("right").isVisible()

    # show curve on right axis
    curve.setVisible(True)
    qapp.processEvents()

    assert plotWidget.getYAxis("right").isVisible()

    # Move curve to left axis
    curve.setYAxis("left")
    qapp.processEvents()

    assert not plotWidget.getYAxis("right").isVisible()


def testAxisSetScaleLogNoData(qapp, qWidgetFactory):
    """Test Axis.setScale('log') method with an empty plot

    Limits are reset only when negative
    """
    plotWidget = qWidgetFactory(PlotWidget)
    xaxis = plotWidget.getXAxis()
    yaxis = plotWidget.getYAxis()
    y2axis = plotWidget.getYAxis("right")

    xaxis.setLimits(-1.0, 1.0)
    yaxis.setLimits(2.0, 3.0)
    y2axis.setLimits(-2.0, -1.0)

    xaxis.setScale("log")
    qapp.processEvents()

    assert xaxis.getLimits() == (1.0, 100.0)
    assert yaxis.getLimits() == (2.0, 3.0)
    assert y2axis.getLimits() == (-2.0, -1.0)

    xaxis.setLimits(10.0, 20.0)

    yaxis.setScale("log")
    qapp.processEvents()

    assert xaxis.getLimits() == (10.0, 20.0)
    assert yaxis.getLimits() == (2.0, 3.0)  # Positive range is preserved
    assert y2axis.getLimits() == (1.0, 100.0)  # Negative min is reset


def testAxisSetScaleLogWithData(qapp, qWidgetFactory):
    """Test Axis.setScale('log') method with data

    Limits are reset only when negative and takes the data range into account
    """
    plotWidget = qWidgetFactory(PlotWidget)
    xaxis = plotWidget.getXAxis()
    yaxis = plotWidget.getYAxis()
    plotWidget.addCurve((-1, 1, 2, 3), (-1, 1, 2, 3))

    xaxis.setLimits(-1.0, 0.5)  # Limits contains no positive data
    yaxis.setLimits(-1.0, 2.0)  # Limits contains positive data

    xaxis.setScale("log")
    yaxis.setScale("log")
    qapp.processEvents()

    assert xaxis.getLimits() == (1.0, 3.0)  # Reset to positive data range
    assert yaxis.getLimits() == (1.0, 2.0)  # Keep max limit


def testAxisSetScaleLinear(qapp, qWidgetFactory):
    """Test Axis.setScale('linear') method: Limits are not changed"""
    plotWidget = qWidgetFactory(PlotWidget)
    xaxis = plotWidget.getXAxis()
    yaxis = plotWidget.getYAxis()
    y2axis = plotWidget.getYAxis("right")
    xaxis.setScale("log")
    yaxis.setScale("log")
    plotWidget.resetZoom()
    qapp.processEvents()

    xaxis.setLimits(10.0, 1000.0)
    yaxis.setLimits(20.0, 2000.0)
    y2axis.setLimits(30.0, 3000.0)

    xaxis.setScale("linear")
    qapp.processEvents()

    assert xaxis.getLimits() == (10.0, 1000.0)
    assert yaxis.getLimits() == (20.0, 2000.0)
    assert y2axis.getLimits() == (30.0, 3000.0)

    yaxis.setScale("linear")
    qapp.processEvents()

    assert xaxis.getLimits() == (10.0, 1000.0)
    assert yaxis.getLimits() == (20.0, 2000.0)
    assert y2axis.getLimits() == (30.0, 3000.0)



# Adapted from https://github.com/spyder-ide/qtpy/blob/296dee3da8aba381b3cf17da34a6d17626e50357/qtpy/uic.py
# In PySide, loadUi does not exist, so we define it using QUiLoader, and
# then make sure we expose that function. This is adapted from qt-helpers
# which was released under a 3-clause BSD license:
# qt-helpers - a common front-end to various Qt modules
#
# Copyright (c) 2015, Chris Beaumont and Thomas Robitaille
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the
#    distribution.
#  * Neither the name of the Glue project nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Which itself was based on the solution at
#
# https://gist.github.com/cpbotha/1b42a20c8f3eb9bb7cb8
#
# which was released under the MIT license:
#
# Copyright (c) 2011 Sebastian Wiesner <lunaryorn@gmail.com>
# Modifications by Charl Botha <cpbotha@vxlabs.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

"""How to load a user interface dynamically with PySide6"""

import logging

from ._qt import BINDING

if BINDING != "PySide6":
    raise RuntimeError("Unsupported Qt binding: %s", BINDING)

from PySide6.QtCore import QMetaObject, Property, Qt
from PySide6.QtWidgets import QFrame
from PySide6.QtUiTools import QUiLoader


_logger = logging.getLogger(__name__)


# Specific custom widgets


class _Line(QFrame):
    """Widget to use as 'Line' Qt designer"""

    def __init__(self, parent=None):
        super(_Line, self).__init__(parent)
        self.setFrameShape(QFrame.HLine)
        self.setFrameShadow(QFrame.Sunken)

    def getOrientation(self):
        shape = self.frameShape()
        if shape == QFrame.HLine:
            return Qt.Horizontal
        elif shape == QFrame.VLine:
            return Qt.Vertical
        else:
            raise RuntimeError("Wrong shape: %d", shape)

    def setOrientation(self, orientation):
        if orientation == Qt.Horizontal:
            self.setFrameShape(QFrame.HLine)
        elif orientation == Qt.Vertical:
            self.setFrameShape(QFrame.VLine)
        else:
            raise ValueError("Unsupported orientation %s" % str(orientation))

    orientation = Property("Qt::Orientation", getOrientation, setOrientation)


CUSTOM_WIDGETS = {"Line": _Line}
"""Default custom widgets for `loadUi`"""


class UiLoader(QUiLoader):
    """
    Subclass of :class:`~PySide.QtUiTools.QUiLoader` to create the user
    interface in a base instance.

    Unlike :class:`~PySide.QtUiTools.QUiLoader` itself this class does not
    create a new instance of the top-level widget, but creates the user
    interface in an existing instance of the top-level class if needed.

    This mimics the behaviour of :func:`PyQt4.uic.loadUi`.
    """

    def __init__(self, baseinstance, customWidgets=None):
        """
        Create a loader for the given ``baseinstance``.

        The user interface is created in ``baseinstance``, which must be an
        instance of the top-level class in the user interface to load, or a
        subclass thereof.

        ``customWidgets`` is a dictionary mapping from class name to class
        object for custom widgets. Usually, this should be done by calling
        registerCustomWidget on the QUiLoader, but with PySide 1.1.2 on
        Ubuntu 12.04 x86_64 this causes a segfault.

        ``parent`` is the parent object of this loader.
        """

        QUiLoader.__init__(self, baseinstance)

        self.baseinstance = baseinstance

        if customWidgets is None:
            self.customWidgets = {}
        else:
            self.customWidgets = customWidgets

    def createWidget(self, class_name, parent=None, name=""):
        """
        Function that is called for each widget defined in ui file,
        overridden here to populate baseinstance instead.
        """

        if parent is None and self.baseinstance:
            # supposed to create the top-level widget, return the base
            # instance instead
            return self.baseinstance

        else:
            # For some reason, Line is not in the list of available
            # widgets, but works fine, so we have to special case it here.
            if class_name in self.availableWidgets() or class_name == "Line":
                # create a new widget for child widgets
                widget = QUiLoader.createWidget(self, class_name, parent, name)

            else:
                # If not in the list of availableWidgets, must be a custom
                # widget. This will raise KeyError if the user has not
                # supplied the relevant class_name in the dictionary or if
                # customWidgets is empty.
                try:
                    widget = self.customWidgets[class_name](parent)
                except KeyError as error:
                    raise Exception(
                        f"No custom widget {class_name} " "found in customWidgets"
                    ) from error

            if self.baseinstance:
                # set an attribute for the new child widget on the base
                # instance, just like PyQt4.uic.loadUi does.
                setattr(self.baseinstance, name, widget)

            return widget


def _get_custom_widgets(ui_file):
    """
    This function is used to parse a ui file and look for the <customwidgets>
    section, then automatically load all the custom widget classes.
    """

    import sys
    import importlib
    from xml.etree.ElementTree import ElementTree

    # Parse the UI file
    etree = ElementTree()
    ui = etree.parse(ui_file)

    # Get the customwidgets section
    custom_widgets = ui.find("customwidgets")

    if custom_widgets is None:
        return {}

    custom_widget_classes = {}

    for custom_widget in list(custom_widgets):
        cw_class = custom_widget.find("class").text
        cw_header = custom_widget.find("header").text

        module = importlib.import_module(cw_header)

        custom_widget_classes[cw_class] = getattr(module, cw_class)

    return custom_widget_classes


def loadUi(uifile, baseinstance=None, package=None, resource_suffix=None):
    """
    Dynamically load a user interface from the given ``uifile``.

    ``uifile`` is a string containing a file name of the UI file to load.

    If ``baseinstance`` is ``None``, the a new instance of the top-level
    widget will be created. Otherwise, the user interface is created within
    the given ``baseinstance``. In this case ``baseinstance`` must be an
    instance of the top-level widget class in the UI file to load, or a
    subclass thereof. In other words, if you've created a ``QMainWindow``
    interface in the designer, ``baseinstance`` must be a ``QMainWindow``
    or a subclass thereof, too. You cannot load a ``QMainWindow`` UI file
    with a plain :class:`~PySide.QtGui.QWidget` as ``baseinstance``.

    :method:`~PySide.QtCore.QMetaObject.connectSlotsByName()` is called on
    the created user interface, so you can implemented your slots according
    to its conventions in your widget class.

    Return ``baseinstance``, if ``baseinstance`` is not ``None``. Otherwise
    return the newly created instance of the user interface.
    """
    if package is not None:
        _logger.warning("loadUi package parameter not implemented with PySide")
    if resource_suffix is not None:
        _logger.warning("loadUi resource_suffix parameter not implemented with PySide")

    # We parse the UI file and import any required custom widgets
    customWidgets = _get_custom_widgets(uifile)

    # Add CUSTOM_WIDGETS
    for name, klass in CUSTOM_WIDGETS.items():
        customWidgets.setdefault(name, klass)

    loader = UiLoader(baseinstance, customWidgets)

    widget = loader.load(uifile)
    QMetaObject.connectSlotsByName(widget)
    return widget



#!/usr/bin/env python3
# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Build man pages of the project's entry points

The project's package MUST be installed in the current Python environment.
"""

import logging
from pathlib import Path
import re
import subprocess
import sys
from typing import Iterator, Tuple

from setuptools.config import read_configuration


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


PROJECT_PATH = Path(__file__).parent.parent


def entry_points(project_path: Path) -> Iterator[Tuple[str, str, str]]:
    config = read_configuration(project_path / "setup.cfg")
    entry_points_config = config.get("options", {}).get("entry_points", {})
    print(entry_points_config)

    for group in ("console_scripts", "gui_scripts"):
        for entry in entry_points_config.get(group, ()):
            print(entry)
            match = re.fullmatch(
                r"(?P<name>\w+)\s*=\s*(?P<module>[^:]+):(?P<object>\w*)",
                entry,
            )
            if match is not None:
                yield match["name"], match["module"], match["object"]


def get_synopsis(module_name: str) -> str:
    """Execute Python commands to retrieve the synopsis for help2man"""
    commands = (
        "import logging",
        "logging.basicConfig(level=logging.ERROR)",
        f"import {module_name}",
        f"print({module_name}.__doc__)",
    )
    result = subprocess.run(
        [sys.executable, "-c", "; ".join(commands)],
        capture_output=True,
    )
    if result.returncode:
        logger.warning("Error while getting synopsis for module '%s'.", module_name)
        return None
    synopsis = result.stdout.decode("utf-8").strip()
    if synopsis == "None":
        return None
    return synopsis


def main(project_path: Path, out_path: Path):
    out_path.mkdir(parents=True, exist_ok=True)

    eps = tuple(entry_points(project_path))
    if not eps:
        raise RuntimeError("No entry points found!")

    for target_name, module_name, function_name in eps:
        logger.info(f"Build man for entry-point target '{target_name}'")
        python_command = [
            sys.executable,
            "-c",
            f'"import {module_name}; {module_name}.{function_name}()"',
        ]

        help2man_command = [
            "help2man",
            "-N",
            " ".join(python_command),
            "-o",
            str(out_path / f"{target_name}.1"),
        ]

        synopsis = get_synopsis(module_name)
        if synopsis:
            help2man_command += ["-n", synopsis]

        result = subprocess.run(help2man_command)
        if result.returncode != 0:
            logger.error(f"Error while generating man file for target '{target_name}'.")
            for argument in ("--help", "--version"):
                test_command = python_command + [argument]
                logger.info(f"Running: {test_command}")
                result = subprocess.run(test_command)
                logger.info(f"\tReturn code: {result.returncode}")
            raise RuntimeError(f"Fail to generate '{target_name}' man documentation")


if __name__ == "__main__":
    main(project_path=PROJECT_PATH, out_path=PROJECT_PATH / "build" / "man")



# /*##########################################################################
#
# Copyright (c) 2017-2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module provides the :class:`Curve` item of the :class:`Plot`.
"""
from __future__ import annotations

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "24/04/2018"


import logging

import numpy

from ....utils.deprecation import deprecated_warning
from ... import colors
from .core import (
    PointsBase,
    LabelsMixIn,
    ColorMixIn,
    YAxisMixIn,
    FillMixIn,
    LineMixIn,
    LineGapColorMixIn,
    LineStyleType,
    SymbolMixIn,
    BaselineMixIn,
    HighlightedMixIn,
    _Style,
)
from silx._utils import NP_OPTIONAL_COPY


_logger = logging.getLogger(__name__)


class CurveStyle(_Style):
    """Object storing the style of a curve.

    Set a value to None to use the default

    :param color: Color
    :param linestyle: Style of the line
    :param linewidth: Width of the line
    :param symbol: Symbol for markers
    :param symbolsize: Size of the markers
    :param gapcolor: Color of gaps of dashed line
    """

    def __init__(
        self,
        color: colors.RGBAColorType | None = None,
        linestyle: LineStyleType | None = None,
        linewidth: float | None = None,
        symbol: str | None = None,
        symbolsize: float | None = None,
        gapcolor: colors.RGBAColorType | None = None,
    ):
        if color is None:
            self._color = None
        else:
            if isinstance(color, str):
                color = colors.rgba(color)
            else:  # array-like expected
                color = numpy.asarray(color)
                if color.ndim == 1:  # Array is 1D, this is a single color
                    color = colors.rgba(color)
            self._color = color

        if not LineMixIn.isValidLineStyle(linestyle):
            raise ValueError(f"Not a valid line style: {linestyle}")
        self._linestyle = linestyle

        self._linewidth = None if linewidth is None else float(linewidth)

        if symbol is not None:
            assert symbol in SymbolMixIn.getSupportedSymbols()
        self._symbol = symbol

        self._symbolsize = None if symbolsize is None else float(symbolsize)

        self._gapcolor = None if gapcolor is None else colors.rgba(gapcolor)

    def getColor(self, copy=True):
        """Returns the color or None if not set.

        :param bool copy: True to get a copy (default),
            False to get internal representation (do not modify!)

        :rtype: Union[List[float],None]
        """
        if isinstance(self._color, numpy.ndarray):
            return numpy.array(self._color, copy=copy or NP_OPTIONAL_COPY)
        else:
            return self._color

    def getLineGapColor(self):
        """Returns the color of dashed line gaps or None if not set.

        :rtype: Union[List[float],None]
        """
        return self._gapcolor

    def getLineStyle(self) -> LineStyleType | None:
        """Return the type of the line or None if not set.

        Type of line::

            - ' '  no line
            - '-'  solid line
            - '--' dashed line
            - '-.' dash-dot line
            - ':'  dotted line
            - (offset, (dash pattern))
        """
        return self._linestyle

    def getLineWidth(self):
        """Return the curve line width in pixels or None if not set.

        :rtype: Union[float,None]
        """
        return self._linewidth

    def getSymbol(self):
        """Return the point marker type.

        Marker type::

            - 'o' circle
            - '.' point
            - ',' pixel
            - '+' cross
            - 'x' x-cross
            - 'd' diamond
            - 's' square

        :rtype: Union[str,None]
        """
        return self._symbol

    def getSymbolSize(self):
        """Return the point marker size in points.

        :rtype: Union[float,None]
        """
        return self._symbolsize

    def __eq__(self, other):
        if isinstance(other, CurveStyle):
            return (
                numpy.array_equal(self.getColor(), other.getColor())
                and self.getLineStyle() == other.getLineStyle()
                and self.getLineWidth() == other.getLineWidth()
                and self.getSymbol() == other.getSymbol()
                and self.getSymbolSize() == other.getSymbolSize()
                and self.getLineGapColor() == other.getLineGapColor()
            )
        else:
            return False


class Curve(
    PointsBase,
    ColorMixIn,
    YAxisMixIn,
    FillMixIn,
    LabelsMixIn,
    LineMixIn,
    LineGapColorMixIn,
    BaselineMixIn,
    HighlightedMixIn,
):
    """Description of a curve"""

    _DEFAULT_Z_LAYER = 1
    """Default overlay layer for curves"""

    _DEFAULT_SELECTABLE = True
    """Default selectable state for curves"""

    _DEFAULT_LINEWIDTH = 1.0
    """Default line width of the curve"""

    _DEFAULT_LINESTYLE = "-"
    """Default line style of the curve"""

    _DEFAULT_HIGHLIGHT_STYLE = CurveStyle(color="black")
    """Default highlight style of the item"""

    _DEFAULT_BASELINE = None

    def __init__(self):
        PointsBase.__init__(self)
        ColorMixIn.__init__(self)
        YAxisMixIn.__init__(self)
        FillMixIn.__init__(self)
        LabelsMixIn.__init__(self)
        LineMixIn.__init__(self)
        LineGapColorMixIn.__init__(self)
        BaselineMixIn.__init__(self)
        HighlightedMixIn.__init__(self)

        self._setBaseline(Curve._DEFAULT_BASELINE)

    def _addBackendRenderer(self, backend):
        """Update backend renderer"""
        # Filter-out values <= 0
        xFiltered, yFiltered, xerror, yerror = self.getData(copy=False, displayed=True)

        if len(xFiltered) == 0 or not numpy.any(numpy.isfinite(xFiltered)):
            return None  # No data to display, do not add renderer to backend

        style = self.getCurrentStyle()

        return backend.addCurve(
            xFiltered,
            yFiltered,
            color=style.getColor(),
            gapcolor=style.getLineGapColor(),
            symbol=style.getSymbol(),
            linestyle=style.getLineStyle(),
            linewidth=style.getLineWidth(),
            yaxis=self.getYAxis(),
            xerror=xerror,
            yerror=yerror,
            fill=self.isFill(),
            alpha=self.getAlpha(),
            symbolsize=style.getSymbolSize(),
            baseline=self.getBaseline(copy=False),
        )

    def __getitem__(self, item):
        """Compatibility with PyMca and silx <= 0.4.0"""
        deprecated_warning(
            "Attributes",
            "__getitem__",
            since_version="2.0.0",
            replacement="Use Curve methods",
        )
        if isinstance(item, slice):
            return [self[index] for index in range(*item.indices(5))]
        elif item == 0:
            return self.getXData(copy=False)
        elif item == 1:
            return self.getYData(copy=False)
        elif item == 2:
            return self.getName()
        elif item == 3:
            info = self.getInfo(copy=False)
            return {} if info is None else info
        elif item == 4:
            params = {
                "info": self.getInfo(),
                "color": self.getColor(),
                "symbol": self.getSymbol(),
                "linewidth": self.getLineWidth(),
                "linestyle": self.getLineStyle(),
                "xlabel": self.getXLabel(),
                "ylabel": self.getYLabel(),
                "yaxis": self.getYAxis(),
                "xerror": self.getXErrorData(copy=False),
                "yerror": self.getYErrorData(copy=False),
                "z": self.getZValue(),
                "selectable": self.isSelectable(),
                "fill": self.isFill(),
            }
            return params
        else:
            raise IndexError("Index out of range: %s", str(item))

    def getCurrentStyle(self):
        """Returns the current curve style.

        Curve style depends on curve highlighting

        :rtype: CurveStyle
        """
        if self.isHighlighted():
            style = self.getHighlightedStyle()
            color = style.getColor()
            linestyle = style.getLineStyle()
            linewidth = style.getLineWidth()
            symbol = style.getSymbol()
            symbolsize = style.getSymbolSize()
            gapcolor = style.getLineGapColor()

            return CurveStyle(
                color=self.getColor() if color is None else color,
                linestyle=self.getLineStyle() if linestyle is None else linestyle,
                linewidth=self.getLineWidth() if linewidth is None else linewidth,
                symbol=self.getSymbol() if symbol is None else symbol,
                symbolsize=self.getSymbolSize() if symbolsize is None else symbolsize,
                gapcolor=self.getLineGapColor() if gapcolor is None else gapcolor,
            )

        else:
            return CurveStyle(
                color=self.getColor(),
                linestyle=self.getLineStyle(),
                linewidth=self.getLineWidth(),
                symbol=self.getSymbol(),
                symbolsize=self.getSymbolSize(),
                gapcolor=self.getLineGapColor(),
            )

    def setData(self, x, y, xerror=None, yerror=None, baseline=None, copy=True):
        """Set the data of the curve.

        :param numpy.ndarray x: The data corresponding to the x coordinates.
        :param numpy.ndarray y: The data corresponding to the y coordinates.
        :param xerror: Values with the uncertainties on the x values
        :type xerror: A float, or a numpy.ndarray of float32.
                      If it is an array, it can either be a 1D array of
                      same length as the data or a 2D array with 2 rows
                      of same length as the data: row 0 for positive errors,
                      row 1 for negative errors.
        :param yerror: Values with the uncertainties on the y values.
        :type yerror: A float, or a numpy.ndarray of float32. See xerror.
        :param baseline: curve baseline
        :type baseline: Union[None,float,numpy.ndarray]
        :param bool copy: True make a copy of the data (default),
                          False to use provided arrays.
        """
        PointsBase.setData(self, x=x, y=y, xerror=xerror, yerror=yerror, copy=copy)
        self._setBaseline(baseline=baseline)



# /*##########################################################################
#
# Copyright (c) 2018-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module provides main objects shared by the compare image plot.
"""

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "09/06/2023"


import numpy
import enum
import contextlib
from typing import NamedTuple

from silx.gui.plot.items.image import ImageBase
from silx.gui.plot.items.core import ItemChangedType, ColormapMixIn

from silx._utils import NP_OPTIONAL_COPY
from silx.opencl import ocl

if ocl is not None:
    try:
        from silx.opencl import sift
    except ImportError:
        # sift module is not available (e.g., in official Debian packages)
        sift = None
else:  # No OpenCL device or no pyopencl
    sift = None


@enum.unique
class VisualizationMode(enum.Enum):
    """Enum for each visualization mode available."""

    ONLY_A = "a"
    ONLY_B = "b"
    VERTICAL_LINE = "vline"
    HORIZONTAL_LINE = "hline"
    COMPOSITE_RED_BLUE_GRAY = "rbgchannel"
    COMPOSITE_RED_BLUE_GRAY_NEG = "rbgnegchannel"
    COMPOSITE_A_MINUS_B = "aminusb"


@enum.unique
class AlignmentMode(enum.Enum):
    """Enum for each alignment mode available."""

    ORIGIN = "origin"
    CENTER = "center"
    STRETCH = "stretch"
    AUTO = "auto"


class AffineTransformation(NamedTuple):
    """Description of a 2D affine transformation: translation, scale and
    rotation.
    """

    tx: float
    ty: float
    sx: float
    sy: float
    rot: float


class _CompareImageItem(ImageBase, ColormapMixIn):
    """Description of a virtual item of images to compare, in order to share
    the data through the silx components.
    """

    def __init__(self):
        ImageBase.__init__(self)
        ColormapMixIn.__init__(self)
        self.__image1 = None
        self.__image2 = None
        self.__vizualisationMode = VisualizationMode.ONLY_A

    def getImageData1(self):
        return self.__image1

    def getImageData2(self):
        return self.__image2

    def setImageData1(self, image1):
        if self.__image1 is image1:
            return
        self.__image1 = image1
        self._updated(ItemChangedType.DATA)

    def setImageData2(self, image2):
        if self.__image2 is image2:
            return
        self.__image2 = image2
        self._updated(ItemChangedType.DATA)

    def getVizualisationMode(self) -> VisualizationMode:
        return self.__vizualisationMode

    @contextlib.contextmanager
    def _updateColormapRange(self, previousMode, mode):
        """COMPOSITE_A_MINUS_B don't have the same data range than others.

        If the colormap is using a fixed range, it is updated in order to set
        a similar range with the new data.
        """
        normalize_colormap = (
            previousMode == VisualizationMode.COMPOSITE_A_MINUS_B
            or mode == VisualizationMode.COMPOSITE_A_MINUS_B
        )
        if normalize_colormap:
            data = self._getConcatenatedData(copy=False)
            if data is None or data.size == 0:
                normalize_colormap = False
            else:
                std1 = numpy.nanstd(data)
                mean1 = numpy.nanmean(data)
        yield

        def transfer(v, std1, mean1, std2, mean2):
            """Transfer a value from a data range to another using statistics"""
            if v is None:
                return None
            rv = (v - mean1) / std1
            return rv * std2 + mean2

        if normalize_colormap:
            data = self._getConcatenatedData(copy=False)
            if data is not None and data.size != 0:
                std2 = numpy.nanstd(data)
                mean2 = numpy.nanmean(data)
                c = self.getColormap()
                if c is not None:
                    vmin, vmax = c.getVRange()
                    vmin = transfer(vmin, std1, mean1, std2, mean2)
                    vmax = transfer(vmax, std1, mean1, std2, mean2)
                    c.setVRange(vmin, vmax)

    def setVizualisationMode(self, mode: VisualizationMode):
        if self.__vizualisationMode == mode:
            return None
        with self._updateColormapRange(self.__vizualisationMode, mode):
            self.__vizualisationMode = mode
        self._updated(ItemChangedType.DATA)

    def _getConcatenatedData(self, copy=True):
        if self.__image1 is None and self.__image2 is None:
            return None
        if self.__image1 is None:
            return numpy.array(self.__image2, copy=copy or NP_OPTIONAL_COPY)
        if self.__image2 is None:
            return numpy.array(self.__image1, copy=copy or NP_OPTIONAL_COPY)

        if self.__vizualisationMode == VisualizationMode.COMPOSITE_A_MINUS_B:
            # In this case the histogram have to be special
            if self.__image1.shape == self.__image2.shape:
                return self.__image1.astype(numpy.float32) - self.__image2.astype(
                    numpy.float32
                )
        else:
            d1 = self.__image1[numpy.isfinite(self.__image1)]
            d2 = self.__image2[numpy.isfinite(self.__image2)]
            return numpy.concatenate((d1, d2))

    def _updated(self, event=None, checkVisibility=True):
        # Synchronizes colormapped data if changed
        if event in (ItemChangedType.DATA, ItemChangedType.MASK):
            data = self._getConcatenatedData(copy=False)
            return self._setColormappedData(data, copy=False)
        super()._updated(event=event, checkVisibility=checkVisibility)

    def getColormappedData(self, copy=True):
        """
        Reimplementation of the `ColormapMixIn.getColormappedData` method.

        This is used to provide a consistent auto scale on the compared images.
        """
        return self._getConcatenatedData(copy=copy)
